from cookiecutter.main import cookiecutter
from dataclasses import dataclass, asdict
import logging
import numpy as np
import os
import re
from rich import print
from typing import Union, Dict

from sfaira.consts.utils import clean_doi, clean_id_str
from sfaira.commands.questionary import sfaira_questionary

log = logging.getLogger(__name__)


@dataclass
class TemplateAttributes:
    create_extra_description: str = ''
    dataloader_type: str = ''
    doi_sfaira_repr: str = ''
    id: str = ''
    id_without_doi: str = ''
    number_of_datasets: str = 1

    sample_fns: Union[str, Dict[str, list]] = ''

    author: Union[str, list] = ''
    default_embedding: str = ''
    doi_preprint: str = ''
    doi_journal: str = ''
    download_url_data: str = ''
    download_url_meta: str = ''
    organism: str = ''
    primary_data: str = ''
    year: int = 0

    layer_counts: str = ''
    layer_processed: str = ''
    layer_spliced_counts: str = ''
    layer_spliced_processed: str = ''
    layer_unspliced_counts: str = ''
    layer_unspliced_processed: str = ''
    layer_velocity: str = ''

    assay_sc: str = ''
    cell_type: str = ''
    development_stage: str = ''
    disease: str = ''
    ethnicity: str = ''
    organ: str = ''
    sample_source: str = ''
    sex: str = ''

    assay_sc_obs_key: str = ''
    cell_type_obs_key: str = ''
    development_stage_obs_key: str = ''
    disease_obs_key: str = ''
    ethnicity_obs_key: str = ''
    organ_obs_key: str = ''
    sample_source_obs_key: str = ''
    sex_obs_key: str = ''

    spatial_x_coord_obs_key: str = ''
    spatial_y_coord_obs_key: str = ''
    spatial_z_coord_obs_key: str = ''
    vdj_vj_1_obs_key_prefix: str = ''
    vdj_vj_2_obs_key_prefix: str = ''
    vdj_vdj_1_obs_key_prefix: str = ''
    vdj_vdj_2_obs_key_prefix: str = ''
    vdj_c_call_obs_key_suffix: str = ''
    vdj_consensus_count_obs_key_suffix: str = ''
    vdj_d_call_obs_key_suffix: str = ''
    vdj_duplicate_count_obs_key_suffix: str = ''
    vdj_j_call_obs_key_suffix: str = ''
    vdj_junction_obs_key_suffix: str = ''
    vdj_junction_aa_obs_key_suffix: str = ''
    vdj_locus_obs_key_suffix: str = ''
    vdj_productive_obs_key_suffix: str = ''
    vdj_v_call_obs_key_suffix: str = ''

    feature_reference: str = ''
    feature_reference_var_key: str = ''
    feature_type: str = ''
    feature_type_var_key: str = ''

    feature_id_var_key: str = ''
    feature_symbol_var_key: str = ''


class DataloaderCreator:

    def __init__(self, path_loader):
        self.WD = os.path.dirname(__file__)
        self.TEMPLATES_PATH = os.path.join(self.WD, "templates")
        self.template_attributes = TemplateAttributes()
        self.out_path = path_loader

    def create_dataloader(self, path_data):
        """
        Prompts and guides the user through a number of possible data loader choices.
        Prompts the user for required attributes which must be present in the data loader.
        Finally creates the specific cookiecutter data loader template.
        """
        self._prompt_dataloader_configuration(path_data)
        self._create_dataloader_template()

    def _prompt_dataloader_configuration(self, path_data):
        """
        Prompts the user for all required attributes for a data loader such as DOI, author, etc.
        """
        print('[bold blue]Dataset structure meta data (yaml::dataset_structure).')
        self.template_attributes.number_of_datasets = sfaira_questionary(function='text',
                                                                         question='Number of datasets',
                                                                         default='1')
        if int(self.template_attributes.number_of_datasets) == 1:
            self.template_attributes.dataloader_type = 'single_dataset'
        else:
            self.template_attributes.dataloader_type = 'multiple_datasets'
        if int(self.template_attributes.number_of_datasets) > 1:
            self.template_attributes.sample_fns = {'fns': []}
            for ds in range(int(self.template_attributes.number_of_datasets)):
                fn = sfaira_questionary(function='text',
                                        question=f'[.sample_fns] Filename for dataset {ds}',
                                        default=f'data_{ds}.h5ad')
                self.template_attributes.sample_fns['fns'].append(fn)

        print('[bold blue]Dataset meta data (yaml::dataset_wise).')
        print('[bold blue]The following meta data queries are on dataset-resolution. '
              'These can also later be modified manually in the .yaml which has the same effect as setting them here.')
        author = sfaira_questionary(function='text',
                                    question='[.author] Leading author "Last name, first name" '
                                             '(e.g.: Einstein, Albert)',
                                    default='')
        self.template_attributes.author = author.split(';') if ';' in author else author
        first_author = author[0] if isinstance(author, list) else author
        try:
            first_author_lastname = first_author.split(',')[0]
        except KeyError:
            print('[bold yellow] First author was not in the expected format. Using full first author for the id.')
            first_author_lastname = first_author

        self.template_attributes.default_embedding = str(sfaira_questionary(
            function='text',
            question='[.default_embedding] Key of default embedding in .obsm:',
            default=''))

        doi = ""
        counter = 0
        while doi == "":
            if counter > 0:
                print('[bold red]You need to supply either a preprint or a journal DOI!'
                      'Use "no_doi_AUTHOR_SOME-NAME" to name data sets that do not have a corresponding publication.')
            doi_preprint_try = 0
            doi_preprint = ""
            while doi_preprint_try == 0 or (
                    doi_preprint != "" and
                    not re.match(r'\b10\.\d+/[\w.]+\b', doi_preprint) and
                    not doi_preprint.startswith("no_doi")
            ):
                if doi_preprint_try > 0:
                    print(f'[bold red]The entered DOI {doi_preprint} is malformed and needs to be an exact DOI! '
                          'You may leave this field empty if you supply doi_journal.')
                doi_preprint = sfaira_questionary(
                    function='text',
                    question='[.doi_preprint] DOI of the preprint publication (e.g.: 10.1000/j.journal.2022.01.001)',
                    default='')
                doi_preprint_try += 1
            self.template_attributes.doi_preprint = doi_preprint
            doi_journal_try = 0
            doi_journal = ""
            while doi_journal_try == 0 or (
                    doi_journal != "" and
                    not re.match(r'\b10\.\d+/[\w.]+\b', doi_journal) and
                    not doi_journal.startswith("no_doi")
            ):
                if doi_journal_try > 0:
                    print(f'[bold red]The entered DOI {doi_preprint} is malformed and needs to be an exact DOI! '
                          'You may leave this field empty if you supply doi_preprint.')
                doi_journal = sfaira_questionary(
                    function='text',
                    question='[.doi_journal] DOI of the journal publication (e.g.: 10.1000/j.journal.2022.01.001)',
                    default='')
                doi_journal_try += 1
            self.template_attributes.doi_journal = doi_journal
            doi = doi_journal if doi_journal != "" else doi_preprint
            counter += 1
        self.template_attributes.doi_sfaira_repr = clean_doi(doi)
        if os.path.exists(self.path_loader):
            print(f"[bold orange]A data loader for this DOI already exists: {self.path_loader}. "
                  f"If you want to write a second data loader for this publication, "
                  f"proceed with this curation workflow, take care that the old one is not overwritten, though. "
                  f"If you were not aware of the other data loader, consider aborting this phase 1a and "
                  f"decide if want delete the old version of this loader "
                  f"or consider completing the previous curation attempt in phase 1b. "
                  )

        download_url_data_try = 0
        download_url_data = ''
        while download_url_data_try == 0 or download_url_data == '':
            if download_url_data_try > 0:
                print('[bold red]You need to supply a download url of the data.')
            download_url_data = sfaira_questionary(
                function='text',
                question='[.download_url_data] URL to download the data',
                default='')
            download_url_data_try += 1
        self.template_attributes.download_url_data = download_url_data
        self.template_attributes.download_url_meta = sfaira_questionary(
            function='text',
            question='[.download_url_meta] URL to download the meta data (only necessary if different from '
                     'download_url_data)',
            default='')

        self.template_attributes.primary_data = str(sfaira_questionary(
            function='confirm',
            question='[.primary_data] Is this primary data?',
            default='Yes'))

        year_try = 0
        year = ""
        while year_try == 0 or year == "":
            if year_try > 0:
                print('[bold red]You need to supply the year of first publication.')
            year = sfaira_questionary(
                function='text',
                question='[.year] Year of first publication (e.g. 2022)',
                default="")
            year_try += 1
        self.template_attributes.year = int(year)

        # Data matrices in the dataset:
        print('[bold blue]Data matrices (yaml::layers).')
        print('[bold blue]An AnnData object may contain multiple data matrices: raw and processed gene expression '
              'counts, or spliced and unspliced count data and velocity estimates, for example. '
              'Minimally, you need to supply either of the matrices "counts" or "processed". '
              'In the following, "*counts" refers to the INTEGER count of alignment events (e.g. transcripts for RNA). '
              'In the following, "*processed" refers to any processing that modifies these counts, for example: '
              'normalization, batch correction, ambient RNA correction. '
              'These items can also later be modified manually in the .yaml which has the same '
              'effect as setting them here.')

        def format_q_mat_key(attr) -> str:
            return f"Layer that contains {attr} (either 'X', 'raw', or a .layers key)"

        main_layer = ""
        counter = 0
        while main_layer == "":
            if counter > 0:
                print('[bold red]You need to supply either a matrix for "counts" or for "processed"!')
            self.template_attributes.layer_counts = sfaira_questionary(
                function='text',
                question=format_q_mat_key("counts"),
                default='')
            self.template_attributes.layer_processed = sfaira_questionary(
                function='text',
                question=format_q_mat_key("processed counts"),
                default='')
            main_layer = self.template_attributes.layer_counts if self.template_attributes.layer_counts != "" else \
                self.template_attributes.layer_processed
            counter += 1
        self.template_attributes.layer_spliced_counts = sfaira_questionary(
            function='text',
            question=format_q_mat_key("spliced counts"),
            default='')
        self.template_attributes.layer_spliced_processed = sfaira_questionary(
            function='text',
            question=format_q_mat_key("processed spliced counts"),
            default='')
        self.template_attributes.layer_unspliced_counts = sfaira_questionary(
            function='text',
            question=format_q_mat_key("unspliced counts"),
            default='')
        self.template_attributes.layer_unspliced_processed = sfaira_questionary(
            function='text',
            question=format_q_mat_key("processed unspliced counts"),
            default='')
        self.template_attributes.layer_velocity = sfaira_questionary(
            function='text',
            question=format_q_mat_key("gene-wise velocities"),
            default='')

        # Meta data that may also be observation-wise:
        print('[bold blue]Cell-wise meta data that are shared across the dataset (yaml::dataset_or_observation_wise).')
        print('[bold blue]The following meta data queries are on a dataset-resolution and should be set '
              'if they do not vary across the cells in that dataset. '
              'Skip the corresponding cell-wise below annotation if you annotate metadata here. '
              'If these meta data vary across cells in a data set, skip them here and annotate them in the next '
              'section. '
              'These items can also later be modified manually in the .yaml which has the same '
              'effect as setting them here. '
              'A lot of these meta data are ontology constrained.'
              'You should input symbols, ie. readable words and not IDs here.'
              'You can look up term symbols here https://www.ebi.ac.uk/ols/index.')

        def format_q_uns_key(k, attr, onto) -> str:
            return f"[.{k}] Dataset-wide {attr} annotation (from {onto})"

        self.template_attributes.assay_sc = sfaira_questionary(
            function='text',
            question=format_q_uns_key("assay_sc", "assay", "EFO"),
            default='')
        self.template_attributes.cell_type = sfaira_questionary(
            function='text',
            question=format_q_uns_key("cell_type", "cell type", "CL (Cell ontology)"),
            default='')
        self.template_attributes.development_stage = sfaira_questionary(
            function='text',
            question=format_q_uns_key("development_stage", "developmental stage", "hsapdv for human, mmusdv for mouse"),
            default='')
        self.template_attributes.disease = sfaira_questionary(
            function='text',
            question=format_q_uns_key("disease", "disease", "MONDO"),
            default='')
        self.template_attributes.ethnicity = sfaira_questionary(
            function='text',
            question=format_q_uns_key("ethnicity", "ethnicity", "HANCESTRO for human, skip for non-human"),
            default='')
        self.template_attributes.organ = sfaira_questionary(
            function='text',
            question=format_q_uns_key("organ", "organ or tissue", "UBERON"),
            default='')

        organism_try = 0
        organism = ''
        while organism_try == 0 or organism == '':
            if organism_try > 0:
                print('[bold red]You need to supply the main organism.')
            organism = sfaira_questionary(
                function='text',
                question=format_q_uns_key("organism", "main organism", "NCBItaxon"),
                default='')
            organism_try += 1
        self.template_attributes.organism = organism

        self.template_attributes.sample_source = sfaira_questionary(
            function='text',
            question='[.sample_source] Dataset-wide sample source annotation (from '
                     '["primary_tissue", "2d_culture", "3d_culture", "tumor"]):',
            default='')
        self.template_attributes.sex = sfaira_questionary(
            function='text',
            question=format_q_uns_key("sex", "sex", "PATO"),
            default='')

        # Encouraged meta data that tend to be in .obs and would require mapping .tsv:
        print('[bold blue]Cell-wise meta data (yaml::dataset_or_observation_wise).')
        print('[bold blue]The following meta data queries are on cell-resolution '
              'and do not need to be set if they were annotated on a dataset-level above. '
              'Skip cell- and dataset-level entries if the corresponding item is not available at all. '
              'These items can also later be modified manually in the .yaml which has the same effect as setting them '
              'here except of that we do additionally check here if you need to run \'sfaira annotate-dataloader\'.')

        def format_q_obs_key(k, attr, onto) -> str:
            return f"[.{k}] Key of {attr} annotation field in .obs (from {onto})"

        def format_warning_double_curation(attr) -> str:
            return f"[bold yellow]WARNING: you set '{attr}' before already, this new entry is ignored. " \
                   "Define this meta data item either data set wide (as before) or cell-specific (here)."

        # assay_sc:
        self.template_attributes.assay_sc_obs_key = sfaira_questionary(
            function='text',
            question=format_q_obs_key("assay_sc_obs_key", "assay", "EFO"),
            default='')
        if self.template_attributes.assay_sc != "" and self.template_attributes.assay_sc_obs_key != "":
            print(format_warning_double_curation('assay_sc'))
        # cell_type:
        self.template_attributes.cell_type_obs_key = sfaira_questionary(
            function='text',
            question=format_q_obs_key("cell_type_obs_key", "cell type", "CL"),
            default='')
        if self.template_attributes.cell_type != "" and self.template_attributes.cell_type_obs_key != "":
            print(format_warning_double_curation('cell type'))
        # development_stage:
        self.template_attributes.development_stage_obs_key = sfaira_questionary(
            function='text',
            question=format_q_obs_key("development_stage_obs_key", "development stage",
                                      "hsapdv for human, mmusdv for mouse"),
            default='')
        if self.template_attributes.development_stage != "" and \
                self.template_attributes.development_stage_obs_key != "":
            print(format_warning_double_curation('development_stage'))
        # disease:
        self.template_attributes.disease_obs_key = sfaira_questionary(
            function='text',
            question=format_q_obs_key("disease_obs_key", "disease", "MONDO"),
            default='')
        if self.template_attributes.disease != "" and self.template_attributes.disease_obs_key != "":
            print(format_warning_double_curation('disease'))
        # ethnicity:
        self.template_attributes.ethnicity_obs_key = sfaira_questionary(
            function='text',
            question=format_q_obs_key("ethnicity_obs_key", "ethnicity", "HANCESTRO for human, skip for non-human"),
            default='')
        if self.template_attributes.ethnicity != "" and self.template_attributes.ethnicity_obs_key != "":
            print(format_warning_double_curation('ethnicity'))
        # organ:
        self.template_attributes.organ_obs_key = sfaira_questionary(
            function='text',
            question=format_q_obs_key("organ_obs_key", "organ or tissue", "UBERON"),
            default='')
        if self.template_attributes.organ != "" and self.template_attributes.organ_obs_key != "":
            print(format_warning_double_curation('organ'))
        # sample_source:
        self.template_attributes.sample_source_obs_key = sfaira_questionary(
            function='text',
            question='[.sample_source_obs_key] Key of sample source annotation field in .obs (from '
                     '["primary_tissue", "2d_culture", "3d_culture", "tumor"]):',
            default='')
        if self.template_attributes.sample_source != "" and self.template_attributes.sample_source_obs_key != "":
            print(format_warning_double_curation('sex'))
        # sex:
        self.template_attributes.sex_obs_key = sfaira_questionary(
            function='text',
            question=format_q_obs_key("sex_obs_key", "sex", "PATO"),
            default='')
        if self.template_attributes.sex != "" and self.template_attributes.sex_obs_key != "":
            print(format_warning_double_curation('sex'))
        requires_annotate = np.any([x != "" for x in [
            self.template_attributes.assay_sc_obs_key,
            self.template_attributes.cell_type_obs_key,
            self.template_attributes.development_stage_obs_key,
            self.template_attributes.disease_obs_key,
            self.template_attributes.ethnicity_obs_key,
            self.template_attributes.organ_obs_key,
            self.template_attributes.sex_obs_key,
        ]])

        # Modality-specific data:
        print('[bold blue]Meta data that is always defined per cell (yaml::observation_wise).')

        def format_q_modality_obs_key(k, modality, attr) -> str:
            return f"[.{k}] {modality}: Key of {attr} annotation field in .obs"

        has_spatial = sfaira_questionary(
            function='confirm',
            question='[.spatial_*] Modality spatial: Does this data set contain spatial coordinates of observations?',
            default='No')
        if has_spatial:
            self.template_attributes.spatial_x_coord_obs_key = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("spatial_x_coord_obs_key", "Spatial", "x coordinate"),
                default='')
            self.template_attributes.spatial_y_coord_obs_key = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("spatial_y_coord_obs_key", "Spatial", "y coordinate"),
                default='')
            self.template_attributes.spatial_z_coord_obs_key = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("spatial_z_coord_obs_key", "Spatial", "z coordinate"),
                default='')

        has_vdj = sfaira_questionary(
            function='confirm',
            question='[.vdj_*] Modality V(D)J: Does this data set contain V(D)J gene reconstructions by observation?',
            default='No')
        if has_vdj:
            print('[bold blue]V(D)J annotation: Meta data definitions are documented here '
                  'https://docs.airr-community.org/en/latest/datarep/rearrangements.html.')
            print('[bold blue]Below, default column names in scirpy are chose, use deafults if you read the V(D)J data '
                  'with scirpy, for example.')
            self.template_attributes.vdj_vj_1_obs_key_prefix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_vj_1_obs_key_prefix",
                                                   "V(D)J", "prefix of first VJ locus/chain"),
                default='IR_VJ_1_')
            self.template_attributes.vdj_vj_2_obs_key_prefix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_vj_2_obs_key_prefix",
                                                   "V(D)J", "prefix of second VJ locus/chain"),
                default='IR_VJ_2_')
            self.template_attributes.vdj_vdj_1_obs_key_prefix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_vdj_1_obs_key_prefix",
                                                   "V(D)J", "prefix of first VDJ locus/chain"),
                default='IR_VDJ_1_')
            self.template_attributes.vdj_vdj_2_obs_key_prefix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_vdj_2_obs_key_prefix",
                                                   "V(D)J", "prefix of second VDJ locus/chain"),
                default='IR_VDJ_2_')
            self.template_attributes.vdj_c_call_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_c_call_obs_key_suffix", "V(D)J", "C gene"),
                default='c_call')
            self.template_attributes.vdj_d_call_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_d_call_obs_key_suffix", "V(D)J", "D gene"),
                default='d_call')
            self.template_attributes.vdj_j_call_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_j_call_obs_key_suffix", "V(D)J", "J gene"),
                default='j_call')
            self.template_attributes.vdj_v_call_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_v_call_obs_key_suffix", "VDJ", "V gene"),
                default='v_call')
            self.template_attributes.vdj_duplicate_count_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_duplicate_count_obs_key_suffix", "V(D)J",
                                                   "number of duplicate UMIs (duplicate count)"),
                default='duplicate_count')
            self.template_attributes.vdj_junction_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_junction_obs_key_suffix", "V(D)J", "junction nt sequence"),
                default='junction')
            self.template_attributes.vdj_junction_aa_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_junction_aa_obs_key_suffix", "V(D)J", "junction aa sequence"),
                default='junction_aa')
            self.template_attributes.vdj_locus_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_locus_obs_key_suffix", "V(D)J",
                                                   "gene locus (e.g elements from IGH, IGK, IGL, TRA, TRB, TRD, or "
                                                   "TRG)"),
                default='locus')
            self.template_attributes.vdj_productive_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_productive_obs_key_suffix", "V(D)J",
                                                   "'Is the gene productive?'"),
                default='productive')
            self.template_attributes.vdj_consensus_count_obs_key_suffix = sfaira_questionary(
                function='text',
                question=format_q_modality_obs_key("vdj_consensus_count_obs_key_suffix", "V(D)J",
                                                   "number of reads contributing to consensus (consensus count)"),
                default='consensus_count')

        print('[bold blue]Feature-wise meta data that may be shared across the dataset '
              '(yaml::dataset_or_feature_wise).')
        print('[bold blue]In some cases, these meta data need to be set per feature, '
              'chose the dataset-wide or feature-wise ("*_var_key") entry accordingly." '
              'These can also later be modified manually in the .yaml which has the same effect as setting them here.')

        self.template_attributes.feature_reference = str(sfaira_questionary(
            function='text',
            question='[.feature_reference] Genome annotation release per dataset (e.g. Homo_sapiens.GRCh38.104)',
            default=''))
        if self.template_attributes.feature_reference == '':
            self.template_attributes.feature_reference_var_key = str(sfaira_questionary(
                function='text',
                question='[.feature_reference_var_key] .var key of genome annotation release per feature'
                         '(e.g. Homo_sapiens.GRCh38.104)',
                default=''))
        else:
            self.template_attributes.feature_reference_var_key = ''
        self.template_attributes.feature_type = sfaira_questionary(
            function='text',
            question='[.feature_type] Feature type of features per dataset (either rna, protein, or peak)',
            default='')
        if self.template_attributes.feature_type == '':
            self.template_attributes.feature_type_var_key = sfaira_questionary(
                function='text',
                question='[.feature_type_var_key] .var key of feature type of features per feature',
                default='')
        else:
            self.template_attributes.feature_type_var_key = ''

        print('[bold blue]Feature-wise meta data (yaml::feature_wise).')
        print('[bold blue]The following meta data queries are on gene-resolution. '
              'These can also later be modified manually in the .yaml which has the same effect as setting them here.')
        feature_id = ""
        counter = 0
        while feature_id == "":
            if counter > 0:
                print('[bold red]You need to supply either a .var key for feature IDs or for symbols!')
            self.template_attributes.feature_id_var_key = sfaira_questionary(
                function='text',
                question='[.feature_id_var_key] Key of feature ID (e.g. ENSEMBL ID) field in .var',
                default='')
            self.template_attributes.feature_symbol_var_key = sfaira_questionary(
                function='text',
                question='[.feature_symbol_var_key] Key of feature symbol (e.g. gene names) field in .var',
                default='index')
            feature_id = self.template_attributes.feature_id_var_key \
                if self.template_attributes.feature_id_var_key != "" else \
                self.template_attributes.feature_symbol_var_key

        def insert_placeholder(id_str):
            return id_str if id_str != "" else "x"

        self.template_attributes.id_without_doi = \
            f'{insert_placeholder(clean_id_str(self.template_attributes.organism))}_' \
            f'{insert_placeholder(clean_id_str(self.template_attributes.organ))}_' \
            f'{insert_placeholder(clean_id_str(str(self.template_attributes.year)))}_' \
            f'{insert_placeholder(clean_id_str(self.template_attributes.assay_sc))}_' \
            f'{insert_placeholder(clean_id_str(first_author_lastname))}_001'
        self.template_attributes.id = f'{self.template_attributes.id_without_doi}_' \
                                      f'{self.template_attributes.doi_sfaira_repr}'

        print('[bold orange]Sfaira butler: "Up next, follow these steps until the next call of the sfaira CLI:"')
        self.action_counter = 1
        print(f'[bold orange]               "{self.action_counter}) Proceed to modify the .yaml and .py files in '
              f'{self.path_loader}"')
        self.action_counter += 1
        self.check_datadir(path_data=path_data)
        if requires_annotate:
            print(f'[bold orange]               "{self.action_counter}) Proceed to phase 2: '
                  'sfaira annotate-dataloader."')
            self.action_counter += 1
        else:
            print(f'[bold orange]               "{self.action_counter}) Proceed to phase 3: '
                  'sfaira finalize-dataloader."')
            print('[bold orange]               "   You can skip phase 2 (sfaira annotate-dataloader), '
                  'unless you set any of the following items in the .yaml during manual editing:"')
            print('[bold orange]               "   assay_sc_obs_key, cell_line_sc_obs_key, cell_type_sc_obs_key, '
                  'development_stage_sc_obs_key, disease_sc_obs_key, ethnicity_sc_obs_key, organ_sc_obs_key, '
                  'organism_sc_obs_key, sex_sc_obs_key."')
            self.action_counter += 1

    @property
    def path_loader(self):
        return os.path.join(self.out_path, self.template_attributes.doi_sfaira_repr)

    def _template_attributes_to_dict(self) -> dict:
        """
        Create a dict from the our Template Structure dataclass
        :return: The dict containing all key-value pairs with non empty values
        """
        return {key: val for key, val in asdict(self.template_attributes).items() if val != ''}

    def _create_dataloader_template(self):
        template_path = os.path.join(self.TEMPLATES_PATH, self.template_attributes.dataloader_type)
        cookiecutter(f'{template_path}',
                     output_dir=self.out_path,
                     no_input=True,
                     overwrite_if_exists=True,
                     extra_context=self._template_attributes_to_dict())

    def check_datadir(self, path_data):
        path_data = os.path.join(path_data, self.template_attributes.doi_sfaira_repr)
        if not os.path.exists(path_data):
            print(f"[bold red]The unmodified downloaded data files were anticipated to lie in {path_data} "
                  f"but this path was not found. "
                  f"Create this directory and move the raw data files there.")
        else:
            print(f'[bold orange]               "{self.action_counter}) Make sure that the unmodified downloaded data '
                  f'files are in  {path_data}."')
            self.action_counter += 1



from sfaira.versions.metadata.base import Ontology, OntologyList, OntologyHierarchical, OntologyObo, \
    OntologyOboCustom, OntologyCl, OntologyHancestro, OntologyUberon, OntologyHsapdv, OntologyMondo, \
    OntologyMmusdv, OntologySex, OntologyEfo, OntologyCellosaurus, OntologyTaxon, OntologyUberonLifecyclestage
from sfaira.versions.metadata.universe import CelltypeUniverse



import anndata
import numpy as np
import pandas as pd
import scipy.sparse
from typing import Dict, List, Union
import yaml


from sfaira.versions.metadata.maps import prepare_ontology_map


def map_freetext_to_ontology(
        queries: Union[str, List[str]],
        onto: str,
        always_return_dict: bool = False,
        anatomical_constraint: Union[str, None] = None,
        choices_for_perfect_match: bool = False,
        include_synonyms: bool = True,
        keep_strategy: bool = False,
        n_suggest: int = 4,
        omit_target_list: list = [],
        threshold_for_partial_matching: float = 90.,
        **kwargs
) -> Union[List[str], Dict[str, List[str]], str]:
    """
    Map free text node name to ontology node names via sfaira cell type matching.

    For details, see also sfaira.versions.metadata.CelltypeUniverse.prepare_celltype_map_fuzzy()

    :param queries: Free text node label which is to be matched to ontology nodes.
        Can also be a list of strings to query.
    :param onto: Name of ontology to map to.
    :param always_return_dict: Also return a dictionary over queries if only one query was given.
    :param anatomical_constraint: Whether to require suggestions to be within a target anatomy defined within UBERON.
    :param choices_for_perfect_match: Whether to give additional matches if a perfect match was found. Note that there
        are cases in which an apparent perfect match corresponds to a general term which could be specified knowing the
        anatomic location of the sample. If this is False and a perfect match is found, only this perfect match is
        returned as a string, rather than as a list.
    :param include_synonyms: Whether to include synonyms of nodes in string search.
    :param keep_strategy: Whether to keep search results structured by search strategy.
        For details, see also sfaira.versions.metadata.CelltypeUniverse.prepare_celltype_map_fuzzy()
    :param n_suggest: Number of cell types to suggest per search strategy.
    :param omit_target_list: Ontology nodes to not match to.
    :param threshold_for_partial_matching: Maximum fuzzy match score below which lenient matching (ratio) is
        extended through partial_ratio.
    :param kwargs: Additional parameters to CelltypeUniverse.
    :return: List over queries, each entry is:
        A list of high priority matches or perfect match (see choices_for_perfect_match) or, if keep_strategy,
        dictionary of lists of search strategies named by strategy name. If a search strategy yields perfect matches, it
        does not return a list of strings but just a single string.
    """
    if isinstance(queries, str):
        queries = [queries]
    matches_to_return = {}
    matches = prepare_ontology_map(
        source=queries,
        onto=onto,
        match_only=False,
        include_synonyms=include_synonyms,
        anatomical_constraint=anatomical_constraint,
        choices_for_perfect_match=choices_for_perfect_match,
        omit_list=[],
        omit_target_list=omit_target_list,
        n_suggest=n_suggest,
        threshold_for_partial_matching=threshold_for_partial_matching,
    )
    matches = matches[0]
    # Prepare the output:
    for x, matches_i in zip(queries, matches):
        matches_i = matches_i
        # Flatten list of lists:
        # Flatten dictionary of lists and account for string rather than list entries.
        if len(matches_i.values()) == 1 and isinstance(list(matches_i.values())[0], str):
            matches_flat = list(matches_i.values())[0]
        else:
            matches_flat = []
            for xx in matches_i.values():
                if isinstance(xx, list):
                    matches_flat.extend(xx)
                else:
                    assert isinstance(xx, str)
                    matches_flat.append(xx)
        if not choices_for_perfect_match and x in matches_flat:
            matches_to_return.update({x: x})
        elif keep_strategy:
            matches_to_return.update({x: matches_i})
        else:
            matches_to_return.update({x: matches_flat})
    # Only return a list over queries if more than one query was given.
    if len(queries) == 1 and not always_return_dict:
        return matches_to_return[queries[0]]
    else:
        return matches_to_return


def read_yaml(fn) -> Dict[str, Dict[str, Union[str, int, bool]]]:
    """
    Read data yaml file.

    Matches format names to Dataset attribute names.

    :param fn: YAML file name.
    :return: Dictionary of dictionaries of names of Dataset attributes and their values.

        - "attr": Data set attributes.
        - "meta": Meta information of yaml and representation.
    """
    with open(fn) as f:
        yaml_dict = yaml.safe_load(f)
    attr_dict = {}
    meta_dict = {}
    for k, v in yaml_dict.items():
        if k not in ["dataset_structure", "meta"]:
            attr_dict.update(v)
        else:
            meta_dict.update(v)
    return {"attr": attr_dict, "meta": meta_dict}


def collapse_matrix(adata: anndata.AnnData, var_column: str) -> anndata.AnnData:
    """
    Collapses (sum) features with the same var_name in a provided var column.

    Does not retain .varm if duplicated var_names are found.
    keeps .var column of first occurrence of duplicated variables.

    :param adata: Input anndata instance with potential duplicated var names.
    :param var_column: column name in .var that contains the duplicated features of interest
    :return: Processed anndata instance without duplicated var names.
    """
    old_index = adata.var.index.tolist() if var_column == "index" else adata.var[var_column].tolist()
    new_index = list(np.unique(old_index))
    if len(new_index) < len(old_index):
        idx_map = np.array([np.where(x == np.array(old_index))[0] for x in new_index])
        # Build initial matrix from first match.
        data = adata.X[:, np.array([x[0] for x in idx_map])].copy()
        # Add additional matched (duplicates) on top:
        for i, idx in enumerate(idx_map):
            if len(idx) > 1:
                data[:, i] = data[:, i] + adata.X[:, idx[1:]].sum(axis=1)

        # Remove varm and populate var with first occurrence only:
        obs_names = adata.obs_names
        adata = anndata.AnnData(
            X=data,
            obs=adata.obs,
            obsm=adata.obsm,
            var=adata.var.iloc[[old_index.index(x) for x in new_index]],
            uns=adata.uns
        )
        adata.obs_names = obs_names
    return adata


def subset_adata_genes(x, var, feature_id_var_key, feature_symbol_var_key, subset_ids_ensg, subset_ids_symbol):
    # Convert data matrix to csc matrix
    if isinstance(x, np.ndarray):
        # Change NaN to zero. This occurs for example in concatenation of anndata instances.
        if np.any(np.isnan(x)):
            x[np.isnan(x)] = 0
        x = scipy.sparse.csc_matrix(x)
    elif isinstance(x, scipy.sparse.spmatrix):
        x = x.tocsc()
    else:
        raise ValueError(f"Data type {type(x)} not recognized.")

    # Remove unmapped genes
    data_ids_ensg = var.index.values if feature_id_var_key == "index" else var[feature_id_var_key].values
    idx_feature_kept = np.where([x.upper() in subset_ids_ensg for x in data_ids_ensg])[0]
    data_ids_kept = data_ids_ensg[idx_feature_kept]
    x = x[:, idx_feature_kept]
    # Build map of subset_ids to features in x:
    idx_feature_map = np.array([subset_ids_ensg.index(x) for x in data_ids_kept])
    # Create reordered feature matrix based on reference and convert to csr
    x_new = scipy.sparse.csc_matrix((x.shape[0], len(subset_ids_ensg)), dtype=x.dtype)
    # copying this over to the new matrix in chunks of size `steps` prevents a strange scipy error:
    # ... scipy/sparse/compressed.py", line 922, in _zero_many i, j, offsets)
    # ValueError: could not convert integer scalar
    step = 500
    if step < len(idx_feature_map):
        i = 0
        for i in range(0, len(idx_feature_map), step):
            x_new[:, idx_feature_map[i:i + step]] = x[:, i:i + step]
        x_new[:, idx_feature_map[i + step:]] = x[:, i + step:]
    else:
        x_new[:, idx_feature_map] = x
    x_new = x_new.tocsr()

    # Create new var dataframe
    if feature_symbol_var_key == "index":
        var_index = subset_ids_symbol
        var_data = {feature_id_var_key: subset_ids_ensg}
    elif feature_id_var_key == "index":
        var_index = subset_ids_ensg
        var_data = {feature_symbol_var_key: subset_ids_symbol}
    else:
        var_index = None
        var_data = {feature_symbol_var_key: subset_ids_symbol,
                    feature_id_var_key: subset_ids_ensg}
    var_new = pd.DataFrame(data=var_data, index=var_index)
    return x_new, var_new



import abc
import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None
from sfaira.models.base import BasicModelKeras


class BasicModelEmbedding:

    @abc.abstractmethod
    def predict_reconstructed(self, x, **kwargs):
        pass

    @abc.abstractmethod
    def predict_embedding(self, x, **kwargs):
        pass


class BasicModelKerasEmbedding(BasicModelKeras, BasicModelEmbedding):
    """
    This base class defines model attributes shared across all tf.keras embedding models.
    """

    encoder_model: tf.keras.Model

    def predict_reconstructed(self, x, **kwargs):
        return np.split(self.training_model.predict(x), indices_or_sections=2, axis=1)[0]

    def predict_embedding(self, x, **kwargs):
        return self.encoder_model.predict(x)



from sfaira.data.dataloaders.base import DatasetBase, DatasetGroup, DatasetGroupDirectoryOriented, \
    DatasetSuperGroup
from sfaira.data.store import load_store, \
    StoreSingleFeatureSpace, \
    StoreAnndata, StoreDao, \
    StoreMultipleFeatureSpaceBase, \
    StoresAnndata, StoresDao, StoresH5ad
from . import dataloaders
from .dataloaders import Universe
from .interactive import DatasetInteractive
from . import store
from . import utils



import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None
from typing import List, Union, Tuple

from sfaira.models.embedding.output_layers import NegBinOutput, NegBinSharedDispOutput, NegBinConstDispOutput, \
    GaussianOutput, GaussianSharedStdOutput, GaussianConstStdOutput
from sfaira.versions.topologies import TopologyContainer
from sfaira.models.embedding.base import BasicModelKerasEmbedding
from sfaira.models.pp_layer import PreprocInput


def log_sum_of_exponentials(x, axis):
    x_max = tf.math.reduce_max(x, axis=axis)
    x_max_expanded = tf.expand_dims(x_max, axis=axis)
    return x_max + tf.math.log(tf.reduce_sum(tf.exp(x - x_max_expanded), axis=axis))


def log_normal_diag(x, mean, log_var, average=False, axis=None):
    log2pi = np.log(2 * np.pi)
    log_normal = -0.5 * (log_var + tf.square(x - mean) / tf.exp(log_var) + log2pi)
    if average:
        return tf.mean(log_normal, axis=axis)
    else:
        return tf.reduce_sum(log_normal, axis=axis)


class Sampling(tf.keras.layers.Layer):
    """Uses (z_mean, z_log_var) to sample z."""

    def call(self, inputs, **kwargs):
        z_mean, z_log_var = inputs
        batch = tf.shape(z_mean)[0]
        dim = tf.shape(z_mean)[1]
        epsilon = tf.keras.backend.random_normal(shape=(batch, dim))
        return z_mean + tf.exp(0.5 * z_log_var) * epsilon


class DensityBlock(tf.keras.layers.Layer):
    def __init__(
            self,
            latent_dim
    ):
        super(DensityBlock, self).__init__(name='density_block')

        self.sampling = Sampling()

        self.dense_mean = tf.keras.layers.Dense(
            units=latent_dim,
            activation="linear"
        )
        self.dense_log_var = tf.keras.layers.Dense(
            units=latent_dim,
            activation="linear"
        )

    def call(self, input_tensor):
        mean = self.dense_mean(input_tensor)
        log_var = tf.clip_by_value(
            self.dense_log_var(input_tensor),
            clip_value_min=np.log(0.001),  # variance larger 0.001
            clip_value_max=np.log(100.0)  # variance smaller 100.0
        )
        x = self.sampling((mean, log_var))
        return x, mean, log_var


class DenseBlock(tf.keras.layers.Layer):
    def __init__(
            self,
            config,
            units
    ):
        super(DenseBlock, self).__init__(name='dense_block')
        _, l1_coef, l2_coef, dropout_rate, self.batchnorm, activation, kernel_initializer = config

        self.dense = tf.keras.layers.Dense(
            units=units,
            activation=activation,
            kernel_initializer=kernel_initializer,
            kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef),
        )
        self.bn = tf.keras.layers.BatchNormalization(
            center=True,
            scale=True
        )
        self.dropout = tf.keras.layers.Dropout(
            dropout_rate,
            noise_shape=None,
            seed=None
        )

    def call(self, input_tensor, training=False):
        x = self.dense(input_tensor)
        if self.batchnorm:
            x = self.bn(x, training=training)
        x = self.dropout(x)
        return x


class TrainablePseudoInputs(tf.keras.layers.Layer):
    def __init__(self, out_shape, activation="hard_sigmoid", **kwargs):
        self.out_shape = out_shape
        self.pseudo_inputs = None
        super().__init__(**kwargs)
        self.activation = tf.keras.activations.get(activation)

    def build(self, input_shape):
        self.pseudo_inputs = self.add_weight(
            shape=self.out_shape,
            initializer=tf.random_normal_initializer(mean=-0.05, stddev=0.01),
            dtype=tf.float32,
            name='u'
        )
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        return self.activation(self.pseudo_inputs)

    def compute_output_shape(self, input_shape):
        return self.out_shape


class Encoder(tf.keras.layers.Layer):
    """Maps input to embedding space"""

    def __init__(
            self,
            config,
            name='encoder',
            **kwargs
    ):
        super().__init__(name=name, **kwargs)
        outer_hidden_dim, inner_hidden_dim, (z1_dim, z2_dim), _, _ = config[0]

        self.enc_dense1 = DenseBlock(config, units=outer_hidden_dim)
        self.enc_dense2 = DenseBlock(config, units=inner_hidden_dim)
        self.q_z2 = DensityBlock(z2_dim)

        self.enc_dense3a = DenseBlock(config, units=inner_hidden_dim)
        self.enc_dense3b = DenseBlock(config, units=inner_hidden_dim)
        self.enc_dense4 = DenseBlock(config, units=inner_hidden_dim)
        self.q_z1 = DensityBlock(z1_dim)

    def call(self, inputs, **kwargs):                                   # (batch_size, in_dim)
        x = self.enc_dense1(inputs)                                     # (batch_size, outer_hidden_dim)
        x = self.enc_dense2(x)                                          # (batch_size, inner_hidden_dim)
        z2, q_z2_mean, q_z2_log_var = self.q_z2(x)                      # (batch_size, z2_dim)

        x = tf.concat([self.enc_dense3a(z2), self.enc_dense3b(inputs)], axis=1)  # (batch_size, 2 * inner_hidden_dim)
        x = self.enc_dense4(x)                                                   # (batch_size, inner_hidden_dim)
        z1, q_z1_mean, q_z1_log_var = self.q_z1(x)                               # (batch_size, z1_dim)
        return (z1, q_z1_mean, q_z1_log_var), (z2, q_z2_mean, q_z2_log_var)


class Decoder(Encoder):
    """Maps latent space sample back to output"""

    def __init__(
            self,
            in_dim,
            batch_size_u,
            config,
            name='decoder',
            **kwargs
    ):
        super().__init__(config=config, name=name, **kwargs)
        _, _, (z1_dim, z2_dim), inner_hidden_dim, outer_hidden_dim = config[0]

        self.dec_dense1 = DenseBlock(config, units=inner_hidden_dim)
        self.dec_dense2 = DenseBlock(config, units=inner_hidden_dim)
        self.p_z1 = DensityBlock(z1_dim)

        self.dec_dense3a = DenseBlock(config, units=inner_hidden_dim)
        self.dec_dense3b = DenseBlock(config, units=inner_hidden_dim)

        self.dec_dense4 = DenseBlock(config, units=outer_hidden_dim)

        self.pseudo_inputs_layer = TrainablePseudoInputs(out_shape=(batch_size_u, in_dim))

    def call(self, inputs, **kwargs):
        z1, z2, data_input = inputs                                     # (batch_size, z1_dim/z2_dim/in_dim)

        # p(z_1|z_2)
        x = self.dec_dense1(z2)                                         # (batch_size, inner_hidden_dim)
        x = self.dec_dense2(x)                                          # (batch_size, inner_hidden_dim)
        _, p_z1_mean, p_z1_log_var = self.p_z1(x)                       # (batch_size, z1_dim)

        # the prior p(z_2)
        u = self.pseudo_inputs_layer(data_input)                        # (batch_size_u, in_dim)
        x = self.enc_dense1(u)                                          # (batch_size_u, inner_hidden_dim)
        x = self.enc_dense2(x)                                          # (batch_size_u, inner_hidden_dim)
        _, p_z2_mean, p_z2_log_var = self.q_z2(x)                       # (batch_size_u, z2_dim)

        x = tf.concat([self.dec_dense3a(z1), self.dec_dense3b(z2)], axis=1)  # (batch_size, 2 * inner_hidden_dim)
        out = self.dec_dense4(x)                                             # (batch_size, outer_hidden_dim)

        return (p_z1_mean, p_z1_log_var), (p_z2_mean, p_z2_log_var), out


class ModelKerasVaeVamp(BasicModelKerasEmbedding):

    def predict_reconstructed(self, x: np.ndarray):
        return np.split(self.training_model.predict(x)[0], indices_or_sections=2, axis=1)[0]

    def __init__(
            self,
            in_dim,
            latent_dim=(256, 128, (32, 32), 128, 256),
            dropout_rate=0.1,
            l1_coef: float = 0.,
            l2_coef: float = 0.,
            batch_size_u: int = 500,
            batchnorm: bool = False,
            activation='tanh',
            init='glorot_uniform',
            output_layer="nb"
    ):
        super(ModelKerasVaeVamp, self).__init__()
        config = (
            latent_dim,
            l1_coef,
            l2_coef,
            dropout_rate,
            batchnorm,
            activation,
            init
        )
        inputs_encoder = tf.keras.Input(shape=(in_dim,), name='counts')
        inputs_sf = tf.keras.Input(shape=(1,), name='size_factors')
        inputs_encoder_pp = PreprocInput()(inputs_encoder)
        encoder = Encoder(config)
        decoder = Decoder(in_dim, batch_size_u, config)

        (z1, q_z1_mean, q_z1_log_var), (z2, q_z2_mean, q_z2_log_var) = encoder(inputs_encoder_pp)
        (p_z1_mean, p_z1_log_var), (p_z2_mean, p_z2_log_var), out = decoder((z1, z2, inputs_encoder_pp))

        z2_expanded = tf.expand_dims(z2, axis=1)  # (batch_size, 1, dim_z2)
        p_z2_mean_expanded = tf.expand_dims(p_z2_mean, axis=0)  # (1, batch_size_u, dim_z2)
        p_z2_log_var_expanded = tf.expand_dims(p_z2_log_var, axis=0)  # (1, batch_size_u, dim_z2)

        log_p_z1 = log_normal_diag(z1, p_z1_mean, p_z1_log_var, axis=1)
        log_q_z1 = log_normal_diag(z1, q_z1_mean, q_z1_log_var, axis=1)
        # the prior log_p_z2
        log_p_z2 = log_normal_diag(z2_expanded, p_z2_mean_expanded, p_z2_log_var_expanded, axis=2)
        log_p_z2 = log_sum_of_exponentials(log_p_z2, axis=1)  # marginalize over batch_size_u
        log_q_z2 = log_normal_diag(z2, q_z2_mean, q_z2_log_var, axis=1)

        logqz_x = tf.expand_dims(log_q_z1 + log_q_z2, axis=1)
        logpz = tf.expand_dims(log_p_z1 + log_p_z2, axis=1)

        expected_densities = tf.keras.layers.Concatenate(axis=1, name="kl")([logqz_x, logpz])

        if output_layer == 'nb':
            output_decoder_expfamily = NegBinOutput(original_dim=in_dim)((out, inputs_sf))
        elif output_layer == 'nb_shared_disp':
            output_decoder_expfamily = NegBinSharedDispOutput(original_dim=in_dim)((out, inputs_sf))
        elif output_layer == 'nb_const_disp':
            output_decoder_expfamily = NegBinConstDispOutput(original_dim=in_dim)((out, inputs_sf))
        elif output_layer == 'gaussian':
            output_decoder_expfamily = GaussianOutput(original_dim=in_dim)((out, inputs_sf))
        elif output_layer == 'gaussian_shared_disp':
            output_decoder_expfamily = GaussianSharedStdOutput(original_dim=in_dim)((out, inputs_sf))
        elif output_layer == 'gaussian_const_disp':
            output_decoder_expfamily = GaussianConstStdOutput(original_dim=in_dim)((out, inputs_sf))
        else:
            raise ValueError("tried to access a non-supported output layer %s" % output_layer)
        output_decoder_expfamily_concat = tf.keras.layers.Concatenate(axis=1, name="neg_ll")(output_decoder_expfamily)

        z = tf.keras.layers.Concatenate(axis=1, name="z")([z1, z2])
        z_mean = tf.keras.layers.Concatenate(axis=1, name="z_mean")([q_z1_mean, q_z2_mean])
        z_log_var = tf.keras.layers.Concatenate(axis=1, name="z_log_var")([q_z1_log_var, q_z2_log_var])

        self.encoder_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=[z, z_mean, z_log_var],
            name="encoder_model"
        )
        self.training_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=[output_decoder_expfamily_concat, expected_densities],
            name="autoencoder"
        )

    def predict_embedding(self, x, variational=False):
        if variational:
            return self.encoder_model.predict(x)
        else:
            return self.encoder_model.predict(x)[1]


class ModelVaeVampVersioned(ModelKerasVaeVamp):
    def __init__(
            self,
            topology_container: TopologyContainer,
            override_hyperpar: Union[dict, None] = None
    ):
        hyperpar = topology_container.topology["hyper_parameters"]
        if override_hyperpar is not None:
            for k in list(override_hyperpar.keys()):
                hyperpar[k] = override_hyperpar[k]
        super().__init__(
            in_dim=topology_container.n_var,
            **hyperpar
        )

        print('passed hyperpar: \n', hyperpar)
        self._topology_id = topology_container.topology_id
        self.genome_size = topology_container.n_var
        self.model_class = "embedding"
        self.model_type = topology_container.model_type
        self.hyperparam = dict(
            list(hyperpar.items()) +  # noqa: W504
            [
                ("topology_id", self._topology_id),
                ("genome_size", self.genome_size),
                ("model_class", self.model_class),
                ("model_type", self.model_type)
            ]
        )



import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None
from typing import List, Union, Tuple

from sfaira.models.embedding.output_layers import NegBinOutput, NegBinSharedDispOutput, NegBinConstDispOutput, \
    GaussianOutput, GaussianSharedStdOutput, GaussianConstStdOutput
from sfaira.versions.topologies import TopologyContainer
from sfaira.models.embedding.base import BasicModelKerasEmbedding
from sfaira.models.pp_layer import PreprocInput


class Encoder(tf.keras.layers.Layer):
    """Project input data to latent embedding space"""

    def __init__(
            self,
            latent_dim: Tuple,
            l2_coef: float,
            l1_coef: float,
            hidden_dropout: List[float],
            input_dropout: float,
            batchnorm: bool,
            activation: str,
            kernel_initializer: str,
            name='encoder',
            **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.layer_list = []

        if input_dropout > 0.0:
            self.layer_list.append(tf.keras.layers.Dropout(input_dropout, name='input_dropout'))

        # build layer list
        for i, (hid_size, hid_drop) in enumerate(zip(latent_dim, hidden_dropout)):
            # At the bottleneck layer:
            # 1. a linear activation function is used to not restrict the support of the hidden units
            # Additionally, before and after the bottleneck dense layer:
            # 2. no batch normalisation is used to not scale the variance of inactive units up and the activity of
            # active units down.
            # 3. no dropout is used to not confound the required bottleneck dimension with dropout rate.
            self.layer_list.append(
                tf.keras.layers.Dense(
                    hid_size,
                    activation=activation if i < (len(latent_dim) - 1) else "linear",
                    kernel_initializer=kernel_initializer,
                    kernel_regularizer=tf.keras.regularizers.l1_l2(l1_coef, l2_coef),
                    name='enc%s' % i
                )
            )
            if batchnorm and i < (len(latent_dim) - 2):
                self.layer_list.append(tf.keras.layers.BatchNormalization(center=True, scale=False))
            if hid_drop > 0.0 and i < (len(latent_dim) - 2):
                if activation == "selu":
                    self.layer_list.append(tf.keras.layers.AlphaDropout(hid_drop, name='enc_%s_drop' % i))
                else:
                    self.layer_list.append(tf.keras.layers.Dropout(hid_drop, name='enc_%s_drop' % i))

    def call(self, x, **kwargs):
        for layer in self.layer_list:
            x = layer(x)
        return x


class Decoder(tf.keras.layers.Layer):
    """Recover output from latent embedding"""

    def __init__(
            self,
            latent_dim: Tuple,
            l2_coef: float,
            l1_coef: float,
            hidden_dropout: List[float],
            batchnorm: bool,
            activation: str,
            kernel_initializer: str,
            name='decoder',
            **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.layer_list = []

        # build layer list
        for i, (hid_size, hid_drop) in enumerate(zip(latent_dim, hidden_dropout)):
            self.layer_list.append(
                tf.keras.layers.Dense(
                    hid_size,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    kernel_regularizer=tf.keras.regularizers.l1_l2(l1_coef, l2_coef),
                    name='dec%s' % i
                )
            )
            if batchnorm:
                self.layer_list.append(
                    tf.keras.layers.BatchNormalization(center=True, scale=False)
                )

            if hid_drop > 0.0:
                if activation == "selu":
                    self.layer_list.append(tf.keras.layers.AlphaDropout(hid_drop, name='dec_%s_drop' % i))
                else:
                    self.layer_list.append(tf.keras.layers.Dropout(hid_drop, name='dec_%s_drop' % i))

    def call(self, x, **kwargs):
        for layer in self.layer_list:
            x = layer(x)
        return x


class ModelKerasAe(BasicModelKerasEmbedding):
    """Combines the encoder and decoder into an end-to-end model for training."""
    # Note: Original DCA implementation uses l1_l2 regularisation also on last layer (nb) - missing here
    # Note: Original DCA implementation uses softplus function instead of exponential as dispersion activation

    def __init__(
            self,
            in_dim,
            latent_dim: Tuple = (64, 32, 64),
            l2_coef: float = 0.,
            l1_coef: float = 0.,
            dropout_rate: float = 0.,
            input_dropout: float = 0.,
            batchnorm: bool = True,
            activation='relu',
            init='glorot_uniform',
            output_layer="nb"
    ):
        # Check length of latent dim to divide encoder-decoder stack:
        if len(latent_dim) % 2 == 1:
            n_layers_enc = len(latent_dim) // 2 + 1
        else:
            raise ValueError("len(latent_dim)=%i should be uneven to provide a defined bottleneck" % len(latent_dim))

        self.in_dim = in_dim
        self.latent_dim = latent_dim
        out_dim = in_dim
        self.out_dim = in_dim

        if isinstance(dropout_rate, list):
            assert len(dropout_rate) == len(latent_dim)
        else:
            dropout_rate = [dropout_rate] * len(latent_dim)

        inputs_encoder = tf.keras.Input(shape=(in_dim,), name='counts')
        inputs_sf = tf.keras.Input(shape=(1,), name='size_factors')
        inputs_encoder_pp = PreprocInput()(inputs_encoder)
        output_encoder = Encoder(
            latent_dim=latent_dim[:n_layers_enc],
            l2_coef=l2_coef,
            l1_coef=l1_coef,
            hidden_dropout=dropout_rate[:n_layers_enc],
            input_dropout=input_dropout,
            batchnorm=batchnorm,
            activation=activation,
            kernel_initializer=init
        )(inputs_encoder_pp)

        output_decoder = Decoder(
            latent_dim=latent_dim[n_layers_enc:],
            l2_coef=l2_coef,
            l1_coef=l1_coef,
            hidden_dropout=dropout_rate[n_layers_enc:],
            batchnorm=batchnorm,
            activation=activation,
            kernel_initializer=init
        )(output_encoder)

        if output_layer == 'nb':
            output_decoder_expfamily = NegBinOutput(original_dim=out_dim)((output_decoder, inputs_sf))
        elif output_layer == 'nb_shared_disp':
            output_decoder_expfamily = NegBinSharedDispOutput(original_dim=out_dim)((output_decoder, inputs_sf))
        elif output_layer == 'nb_const_disp':
            output_decoder_expfamily = NegBinConstDispOutput(original_dim=out_dim)((output_decoder, inputs_sf))
        elif output_layer == 'gaussian':
            output_decoder_expfamily = GaussianOutput(original_dim=out_dim)((output_decoder, inputs_sf))
        elif output_layer == 'gaussian_shared_disp':
            output_decoder_expfamily = GaussianSharedStdOutput(original_dim=out_dim)((output_decoder, inputs_sf))
        elif output_layer == 'gaussian_const_disp':
            output_decoder_expfamily = GaussianConstStdOutput(original_dim=out_dim)((output_decoder, inputs_sf))
        else:
            raise ValueError("tried to access a non-supported output layer %s" % output_layer)
        output_decoder_expfamily_concat = tf.keras.layers.Concatenate(axis=1, name="neg_ll")(output_decoder_expfamily)

        self.encoder_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=output_encoder,
            name="encoder_model"
        )
        self.training_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=[output_decoder_expfamily_concat],
            name="autoencoder"
        )


class ModelAeVersioned(ModelKerasAe):
    def __init__(
            self,
            topology_container: TopologyContainer,
            override_hyperpar: Union[dict, None] = None
    ):
        hyperpar = topology_container.topology["hyper_parameters"]
        if override_hyperpar is not None:
            for k in list(override_hyperpar.keys()):
                hyperpar[k] = override_hyperpar[k]
        super().__init__(
            in_dim=topology_container.n_var,
            **hyperpar
        )
        print('passed hyperpar: \n', hyperpar)
        self._topology_id = topology_container.topology_id
        self.genome_size = topology_container.n_var
        self.model_class = "embedding"
        self.model_type = topology_container.model_type
        self.hyperparam = dict(
            list(hyperpar.items()) +  # noqa: W504
            [
                ("topology_id", self._topology_id),
                ("genome_size", self.genome_size),
                ("model_class", self.model_class),
                ("model_type", self.model_type)
            ]
        )



import anndata
import tarfile
import gzip
import scipy.io
import os
import pandas as pd
from sfaira.data import DatasetBase

SAMPLE_FNS = [
    "GSM3308545_NOD_08w_A",
    "GSM3308547_NOD_08w_C",
    "GSM3308548_NOD_14w_A",
    "GSM3308549_NOD_14w_B",
    "GSM3308550_NOD_14w_C",
    "GSM3308551_NOD_16w_A",
    "GSM3308552_NOD_16w_B",
    "GSM3308553_NOD_16w_C"
]


class Dataset(DatasetBase):
    """
    TODO add disease
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE117nnn/GSE117770/suppl/GSE117770_RAW.tar"
        self.download_url_meta = f"private,{self.sample_fn}_annotation.csv"

        self.assay_sc = "10x 3' v2"
        self.author = "Thompson"
        self.doi_journal = "10.1016/j.cmet.2019.01.021"
        self.layer_counts = "X"
        self.organ = "pancreas"
        self.organism = "Mus musculus"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.state_exact = "diabetic"
        self.year = 2019

        self.feature_id_var_key = "ensembl"
        self.feature_symbol_var_key = "names"
        self.feature_type = "rna"
        self.cell_type_obs_key = "celltypes"

        self.set_dataset_id(idx=1)


def load(data_dir, sample_fn, **kwargs):
    with tarfile.open(os.path.join(data_dir, 'GSE117770_RAW.tar')) as tar:
        for member in tar.getmembers():
            if "_matrix.mtx.gz" in member.name and sample_fn in member.name:
                name = "_".join(member.name.split("_")[:-1])
                with gzip.open(tar.extractfile(member), "rb") as mm:
                    x = scipy.io.mmread(mm).T.tocsr()
                obs = pd.read_csv(tar.extractfile(name + "_barcodes.tsv.gz"), compression="gzip", header=None,
                                  sep="\t", index_col=0)
                obs.index.name = None
                var = pd.read_csv(tar.extractfile(name + "_genes.tsv.gz"), compression="gzip", header=None,
                                  sep="\t")
                var.columns = ["ensembl", "names"]
                var.index = var["ensembl"].values
                adata = anndata.AnnData(X=x, obs=obs, var=var)
    adata.var_names_make_unique()
    celltypes = pd.read_csv(os.path.join(data_dir, sample_fn + "_annotation.csv"), index_col=0)
    adata = adata[celltypes.index]
    adata.obs["celltypes"] = celltypes

    return adata



import abc
import anndata
import numpy as np
import os
import pandas as pd
import pytest
from typing import Union

from sfaira import settings
from sfaira.consts import AdataIdsSfaira
from sfaira.data import StoreSingleFeatureSpace, StoreMultipleFeatureSpaceBase, load_store
from sfaira.estimators import EstimatorKeras, EstimatorKerasCelltype, EstimatorKerasEmbedding
from sfaira.versions.genomes.genomes import CustomFeatureContainer
from sfaira.versions.metadata import OntologyOboCustom
from sfaira.versions.topologies import TopologyContainer

from sfaira.unit_tests.data_for_tests.loaders.consts import CELLTYPES, CL_VERSION, ASSEMBLY_HUMAN, ASSEMBLY_MOUSE
from sfaira.unit_tests.data_for_tests.loaders.utils import PrepareData
from sfaira.unit_tests.directories import DIR_TEMP

ADATA_IDS = AdataIdsSfaira()
ASSEMBLY = {
    "Homo sapiens": ASSEMBLY_HUMAN,
    "Mus musculus": ASSEMBLY_MOUSE,
}
GENES = {
    "Homo sapiens": ["ENSG00000000003", "ENSG00000000005"],
    "Mus musculus": ["ENSMUSG00000000003", "ENSMUSG00000000028"],
}
TARGETS = CELLTYPES
TARGET_UNIVERSE = CELLTYPES

ASSAYS = ["10x technology", "Smart-seq2"]


TOPOLOGY_EMBEDDING_MODEL = {
    "model_type": None,
    "input": {
        "genome": None,
        "genes": None,
    },
    "output": {},
    "hyper_parameters": {
        "latent_dim": None,
        "l2_coef": 0.,
        "l1_coef": 0.,
        "output_layer": "nb_const_disp"
    }
}

TOPOLOGY_CELLTYPE_MODEL = {
    "model_type": None,
    "input": {
        "genome": None,
        "genes": None,
    },
    "output": {
        "cl": CL_VERSION.split("_")[0],
        "targets": TARGET_UNIVERSE
    },
    "hyper_parameters": {
        "l1_coef": 0.,
        "l2_coef": 0.,
    }
}


class HelperEstimatorBase:

    adata_ids: AdataIdsSfaira
    data: Union[anndata.AnnData, StoreSingleFeatureSpace, StoreMultipleFeatureSpaceBase]
    tc: TopologyContainer

    def load_adata(self, organism="Homo sapiens", organ=None, match_to_reference=None):
        data = PrepareData()
        dsg = data.prepare_dsg(load=True, match_to_release=match_to_reference)
        dsg.subset(key="doi_journal", values=["no_doi_mock1", "no_doi_mock2", "no_doi_mock3"])
        if organism is not None:
            dsg.subset(key="organism", values=organism)
        if organ is not None:
            dsg.subset(key="organ", values=organ)
        self.adata_ids = dsg.dataset_groups[0]._adata_ids
        self.data = dsg.adata_ls

    def load_store(self, organism="Homo sapiens", organ=None, match_to_reference=None):
        data = PrepareData()
        store_path = data.prepare_store(store_format="dao", match_to_reference=match_to_reference)
        store = load_store(cache_path=store_path, store_format="dao")
        store.subset(attr_key="doi_journal", values=["no_doi_mock1", "no_doi_mock2", "no_doi_mock3"])
        if organism is not None:
            store.subset(attr_key="organism", values=organism)
        if organ is not None:
            store.subset(attr_key="organ", values=organ)
        self.adata_ids = store._adata_ids_sfaira
        self.data = store.stores[organism]

    def load_multistore(self):
        data = PrepareData()
        store_path = data.prepare_store(store_format="dao")
        store = load_store(cache_path=store_path, store_format="dao")
        store.subset(attr_key="doi_journal", values=["no_doi_mock1", "no_doi_mock2", "no_doi_mock3"])
        self.adata_ids = store._adata_ids_sfaira
        assert "Mus musculus" in store.stores.keys(), store.stores.keys()
        assert "Homo sapiens" in store.stores.keys(), store.stores.keys()
        self.data = store


class HelperEstimatorKeras(HelperEstimatorBase):

    data: Union[anndata.AnnData, StoreSingleFeatureSpace]
    estimator: Union[EstimatorKeras]
    model_type: str
    tc: TopologyContainer

    """
    Contains functions _test* to test individual functions and attributes of estimator class.

    TODO for everybody working on this, add one _test* function in here and add it into
    basic_estimator_test(). See _test_call() for an example.
    """

    @abc.abstractmethod
    def init_topology(self, model_type: str, feature_space: str, organism: str):
        pass

    @abc.abstractmethod
    def init_estimator(self, test_split):
        """
        Initialise target estimator as .estimator attribute.
        """
        pass

    def estimator_train(self, test_split, randomized_batch_access):
        self.estimator.train(
            optimizer="adam",
            lr=0.005,
            epochs=2,
            batch_size=4,
            validation_split=0.5,
            test_split=test_split,
            validation_batch_size=4,
            max_validation_steps=1,
            shuffle_buffer_size=None if randomized_batch_access else 10,
            cache_full=False,
            randomized_batch_access=randomized_batch_access,
        )

    @abc.abstractmethod
    def basic_estimator_test(self):
        pass

    def load_estimator(self, model_type, data_type, feature_space, test_split, organism="Homo sapiens"):
        self.init_topology(model_type=model_type, feature_space=feature_space, organism=organism)
        np.random.seed(1)
        if data_type == "adata":
            self.load_adata()
        else:
            self.load_store(organism=organism)
        self.init_estimator(test_split=test_split)

    def fatal_estimator_test(self, model_type, data_type, test_split=0.1, feature_space="small"):
        self.load_estimator(model_type=model_type, data_type=data_type, feature_space=feature_space,
                            test_split=test_split)
        self.estimator_train(test_split=test_split, randomized_batch_access=False)
        self.basic_estimator_test()


class HelperEstimatorKerasEmbedding(HelperEstimatorKeras):

    estimator: EstimatorKerasEmbedding
    model_type: str
    tc: TopologyContainer

    def init_topology(self, model_type: str, feature_space: str, organism: str):
        topology = TOPOLOGY_EMBEDDING_MODEL.copy()
        if feature_space == "full":
            # Read 500 genes (not full protein coding) to compromise between being able to distinguish observations
            # and reducing run time of unit tests.
            tab = pd.read_csv(os.path.join(settings.cachedir_genomes, "_".join(organism.split(" ")).lower(),
                                           ASSEMBLY[organism] + ".csv"))
            genes_full = tab.loc[tab["gene_biotype"].values == "protein_coding", "gene_id"].values[:500].tolist()
            topology["input"]["genes"] = ["ensg", genes_full]
        else:
            topology["input"]["genes"] = ["ensg", GENES[organism]]
        topology["input"]["genome"] = ASSEMBLY[organism]
        topology["model_type"] = model_type
        if model_type == "linear":
            topology["hyper_parameters"]["latent_dim"] = 2
        else:
            topology["hyper_parameters"]["latent_dim"] = (2, 2, 2)
        self.model_type = model_type
        self.tc = TopologyContainer(topology=topology, topology_id="0.1")

    def init_estimator(self, test_split):
        self.estimator = EstimatorKerasEmbedding(
            data=self.data,
            model_dir=DIR_TEMP,
            cache_path=DIR_TEMP,
            model_id="testid",
            model_topology=self.tc
        )
        self.estimator.init_model()
        self.estimator.split_train_val_test(test_split=test_split, val_split=0.1)

    def basic_estimator_test(self):
        _ = self.estimator.evaluate()
        prediction_output = self.estimator.predict()
        prediction_embed = self.estimator.predict_embedding()
        weights = self.estimator.model.training_model.get_weights()
        self.estimator.save_weights_to_cache()
        self.estimator.load_weights_from_cache()
        new_prediction_output = self.estimator.predict()
        new_prediction_embed = self.estimator.predict_embedding()
        new_weights = self.estimator.model.training_model.get_weights()
        for i in range(len(weights)):
            if not np.any(np.isnan(weights[i])):
                assert np.allclose(weights[i], new_weights[i], rtol=1e-6, atol=1e-6)
        if self.model_type != 'vae':
            if not np.any(np.isnan(prediction_output)):
                assert np.allclose(prediction_output, new_prediction_output, rtol=1e-6, atol=1e-6)
                assert np.allclose(prediction_embed, new_prediction_embed, rtol=1e-6, atol=1e-6)


class TestHelperEstimatorKerasCelltype(HelperEstimatorKeras):

    estimator: EstimatorKerasCelltype
    nleaves: int
    model_type: str
    tc: TopologyContainer

    def init_topology(self, model_type: str, feature_space: str, organism: str):
        topology = TOPOLOGY_CELLTYPE_MODEL.copy()
        topology["model_type"] = model_type
        topology["input"]["genome"] = ASSEMBLY[organism]
        topology["input"]["genes"] = ["ensg", GENES[organism]]
        if model_type == "mlp":
            topology["hyper_parameters"]["units"] = (2,)
        self.model_type = model_type
        self.tc = TopologyContainer(topology=topology, topology_id="0.0.1")

    def init_estimator(self, test_split):
        tc = self.tc
        if isinstance(self.data, StoreSingleFeatureSpace):
            # Reset leaves below:
            tc.topology["output"]["targets"] = None
        self.estimator = EstimatorKerasCelltype(
            data=self.data,
            model_dir=DIR_TEMP,
            cache_path=DIR_TEMP,
            model_id="testid",
            model_topology=tc,
        )
        obs_cl = self.estimator.data.checkout(obs_keys=[self.adata_ids.cell_type]).obs[self.adata_ids.cell_type].values
        leaves = self.estimator.celltype_universe.onto_cl.get_effective_leaves(
            x=[x for x in obs_cl if x != self.adata_ids.unknown_metadata_identifier]
        )
        self.nleaves = len(leaves)
        self.estimator.celltype_universe.onto_cl.leaves = leaves
        self.estimator.init_model()
        self.estimator.split_train_val_test(test_split=test_split, val_split=0.1)

    def basic_estimator_test(self):
        _ = self.estimator.evaluate()
        prediction_output = self.estimator.predict()
        assert prediction_output.shape[1] == self.nleaves, prediction_output.shape
        weights = self.estimator.model.training_model.get_weights()
        self.estimator.save_weights_to_cache()
        self.estimator.load_weights_from_cache()
        new_prediction_output = self.estimator.predict()
        new_weights = self.estimator.model.training_model.get_weights()
        for i in range(len(weights)):
            if not np.any(np.isnan(weights[i])):
                assert np.allclose(weights[i], new_weights[i], rtol=1e-6, atol=1e-6)
        if not np.any(np.isnan(prediction_output)):
            assert np.allclose(prediction_output, new_prediction_output, rtol=1e-6, atol=1e-6)


class HelperEstimatorKerasCelltypeCustomObo(TestHelperEstimatorKerasCelltype):

    custom_types = ["MYONTO:01", "MYONTO:02", "MYONTO:03"]

    def init_obo_custom(self) -> OntologyOboCustom:
        return OntologyOboCustom(obo=os.path.join(os.path.dirname(__file__), "custom.obo"))

    def init_genome_custom(self, n_features) -> CustomFeatureContainer:
        return CustomFeatureContainer(
            genome_tab=pd.DataFrame({
                "gene_name": ["dim_" + str(i) for i in range(n_features)],
                "gene_id": ["dim_" + str(i) for i in range(n_features)],
                "gene_biotype": ["embedding" for _ in range(n_features)],
            }),
            organism="Homo sapiens",
        )

    def load_adata(self, organism="Homo sapiens", organ=None):
        data = PrepareData()
        dsg = data.prepare_dsg(load=True)
        dsg.subset(key="doi_journal", values=["no_doi_mock1", "no_doi_mock3", "no_doi_mock3"])
        if organism is not None:
            dsg.subset(key="organism", values=organism)
        if organ is not None:
            dsg.subset(key="organ", values=organ)
        self.adata_ids = dsg.dataset_groups[0]._adata_ids
        # Use mock data loading to generate base line object:
        self.data = dsg.adata
        # - Subset to target feature space size:
        self.data = self.data[:, :self.tc.gc.n_var].copy()
        # - Add in custom cell types:
        self.data.obs[self.adata_ids.cell_type] = [
            self.custom_types[np.random.randint(0, len(self.custom_types))]
            for _ in range(self.data.n_obs)
        ]
        self.data.obs[self.adata_ids.cell_type + self.adata_ids.onto_id_suffix] = \
            self.data.obs[self.adata_ids.cell_type]
        # - Add in custom features:
        self.data.var_names = ["dim_" + str(i) for i in range(self.data.n_vars)]
        self.data.var[self.adata_ids.feature_id] = ["dim_" + str(i) for i in range(self.data.n_vars)]
        self.data.var[self.adata_ids.feature_symbol] = ["dim_" + str(i) for i in range(self.data.n_vars)]

    def init_topology_custom(self, model_type: str, n_features):
        topology = TOPOLOGY_CELLTYPE_MODEL.copy()
        topology["model_type"] = model_type
        topology["input"]["genome"] = "custom"
        topology["input"]["genes"] = ["biotype", "embedding"]
        topology["output"]["cl"] = "custom"
        topology["output"]["targets"] = self.custom_types[1:]
        if model_type == "mlp":
            topology["hyper_parameters"]["units"] = (2,)
        self.model_type = model_type
        self.nleaves = len(topology["output"]["targets"])
        gc = self.init_genome_custom(n_features=n_features)
        self.tc = TopologyContainer(topology=topology, topology_id="0.0.1", custom_genome_constainer=gc)

    def fatal_estimator_test_custom(self):
        self.init_topology_custom(model_type="mlp", n_features=50)
        obo = self.init_obo_custom()
        np.random.seed(1)
        self.load_adata()
        self.estimator = EstimatorKerasCelltype(
            data=self.data,
            model_dir=DIR_TEMP,
            cache_path=DIR_TEMP,
            model_id="testid",
            model_topology=self.tc,
            celltype_ontology=obo,
            remove_unlabeled_cells=False,  # TODO this should not be necessary but all cells are filtered otherwise
        )
        self.estimator.init_model()
        self.estimator_train(test_split=0.1, randomized_batch_access=False)
        self.basic_estimator_test()

# Test embedding models:


@pytest.mark.parametrize("data_type", ["adata", "store"])
def test_for_fatal_linear(data_type):
    test_estim = HelperEstimatorKerasEmbedding()
    test_estim.fatal_estimator_test(model_type="linear", data_type=data_type)


@pytest.mark.parametrize("data_type", ["store"])
def test_for_fatal_ae(data_type):
    test_estim = HelperEstimatorKerasEmbedding()
    test_estim.fatal_estimator_test(model_type="ae", data_type=data_type)


@pytest.mark.parametrize("data_type", ["store"])
def test_for_fatal_vae(data_type):
    test_estim = HelperEstimatorKerasEmbedding()
    test_estim.fatal_estimator_test(model_type="vae", data_type=data_type)


# Test cell type predictor models:


@pytest.mark.parametrize("data_type", ["adata", "store"])
def test_for_fatal_mlp(data_type):
    test_estim = TestHelperEstimatorKerasCelltype()
    test_estim.fatal_estimator_test(model_type="mlp", data_type=data_type)


@pytest.mark.parametrize("data_type", ["store"])
def test_for_fatal_marker(data_type):
    test_estim = TestHelperEstimatorKerasCelltype()
    test_estim.fatal_estimator_test(model_type="marker", data_type=data_type)


def test_for_fatal_mlp_custom():
    test_estim = HelperEstimatorKerasCelltypeCustomObo()
    test_estim.fatal_estimator_test_custom()

# Test index sets


@pytest.mark.parametrize("batch_size", [1024, 2048, 4096])
@pytest.mark.parametrize("randomized_batch_access", [False, True])
def test_dataset_size(batch_size: int, randomized_batch_access: bool):
    """
    Test that tf data set from estimator has same size as generator invoked directly from store based on number of
    observations in emitted batches.

    Tests for batch sizes smaller, equal to and larger than retrieval batch size and with and without randomized
    batch access.
    """
    test_estim = HelperEstimatorKerasEmbedding()
    retrieval_batch_size = 2048
    # Need full feature space here because observations are not necessarily different in small model testing feature
    # space with only two genes:
    test_estim.load_estimator(model_type="linear", data_type="store", feature_space="reduced", test_split=0.2,
                              organism="Homo sapiens")
    idx_train = test_estim.estimator.idx_train
    ds_train = test_estim.estimator.get_one_time_tf_dataset(idx=idx_train, batch_size=batch_size, mode='eval')
    x_train_shape = 0
    for x, _ in ds_train.as_numpy_iterator():
        x_train_shape += x[0].shape[0]
    # Define raw store generator on train data to compare and check that it has the same size as tf generator exposed
    # by estimator:

    def map_fn(x, obs):
        return (x, ),

    g_train = test_estim.estimator.data.checkout(idx=idx_train, retrieval_batch_size=retrieval_batch_size,
                                                 randomized_batch_access=randomized_batch_access, map_fn=map_fn)
    x_train2_shape = 0
    for x, in g_train.iterator():
        x = x[0]
        if len(x.shape) == 1:
            x = np.expand_dims(x, axis=0)
        x_train2_shape += x.shape[0]
    assert x_train_shape == x_train2_shape
    assert x_train_shape == len(idx_train)


@pytest.mark.parametrize("data_type", ["adata", "store"])
@pytest.mark.parametrize("test_split", [0.3, {"id": "homosapiens_lung_2021_10xtechnology_mock1_001_no_doi_mock1"}])
def test_split_index_sets(data_type: str, test_split):
    """
    Test that train, val, test split index sets are correct:

        1) complete
        2) non-overlapping
        3) that test indices map to all (float split) or distinct (attribute split) data sets
        4) do not contain duplicated observations within and across splits (defined based on the feature vectors)
    """
    test_estim = HelperEstimatorKerasEmbedding()
    # Need full feature space here because observations are not necessarily different in small model testing feature
    # space with only two genes:
    test_estim.load_estimator(model_type="linear", data_type=data_type, feature_space="full",
                              organism="Homo sapiens", test_split=test_split)
    idx_train = test_estim.estimator.idx_train
    idx_eval = test_estim.estimator.idx_eval
    idx_test = test_estim.estimator.idx_test
    # 1) Assert that index assignment sets sum up to full data set:
    # Make sure that there are no repeated indices in each set.
    assert len(idx_train) == len(np.unique(idx_train))
    assert len(idx_eval) == len(np.unique(idx_eval))
    assert len(idx_test) == len(np.unique(idx_test))
    assert len(idx_train) + len(idx_eval) + len(idx_test) == test_estim.estimator.data.n_obs, \
        (len(idx_train), len(idx_eval), len(idx_test), test_estim.estimator.data.n_obs)
    if isinstance(test_estim.data, StoreSingleFeatureSpace):
        assert np.sum([v.shape[0] for v in test_estim.data.indices.values()]) == test_estim.estimator.data.n_obs
    # 2) Assert that index assignments are exclusive to each split:
    assert len(set(idx_train).intersection(set(idx_eval))) == 0
    assert len(set(idx_train).intersection(set(idx_test))) == 0
    assert len(set(idx_test).intersection(set(idx_eval))) == 0
    # 3) Check partition of index vectors over store data sets matches test split scenario:
    if isinstance(test_estim.estimator.data, StoreSingleFeatureSpace):
        # Prepare data set-wise index vectors that are numbered in the same way as global split index vectors.
        idx_raw = []
        counter = 0
        for v in test_estim.estimator.data.indices.values():
            idx_raw.append(np.arange(counter, counter + len(v)))
            counter += len(v)
        if isinstance(test_split, float):
            # Make sure that indices from each split are in each data set:
            for i, z in enumerate([idx_train, idx_eval, idx_test]):
                matches = [  # in each data set
                    np.any([y in z for y in x])  # at least one match of data set to split index set
                    for x in idx_raw
                ]
                assert np.all(matches), (i, matches)
        else:
            # Make sure that indices from (train, val) and test split are exclusive:
            datasets_train = np.where([  # in each data set
                np.any([y in idx_train for y in x])  # at least one match of data set to split index set
                for x in idx_raw
            ])[0]
            datasets_eval = np.where([  # in each data set
                np.any([y in idx_eval for y in x])  # at least one match of data set to split index set
                for x in idx_raw
            ])[0]
            datasets_test = np.where([  # in each data set
                np.any([y in idx_test for y in x])  # at least one match of data set to split index set
                for x in idx_raw
            ])[0]
            assert np.all(datasets_train == datasets_eval), (datasets_train, datasets_eval, datasets_test)
            assert len(set(datasets_train).intersection(set(datasets_test))) == 0, (datasets_train, datasets_test)
    # 4) Assert that observations mapped to indices are actually unique based on expression vectors:
    # Build numpy arrays of expression input data sets from tensorflow data sets directly from estimator.
    # These data sets are the most processed transformation of the data and stand directly in concat with the model.
    ds_train = test_estim.estimator.get_one_time_tf_dataset(idx=idx_train, batch_size=1024, mode='eval')
    ds_eval = test_estim.estimator.get_one_time_tf_dataset(idx=idx_eval, batch_size=1024, mode='eval')
    ds_test = test_estim.estimator.get_one_time_tf_dataset(idx=idx_test, batch_size=1024, mode='eval')
    # Create two copies of test data set to make sure that re-instantiation of a subset does not cause issues.
    ds_test2 = test_estim.estimator.get_one_time_tf_dataset(idx=idx_test, batch_size=1024, mode='eval')
    x_train = []
    x_eval = []
    x_test = []
    x_test2_shape = 0
    for x, _ in ds_train.as_numpy_iterator():
        x_train.append(x[0])
    x_train = np.concatenate(x_train, axis=0)
    for x, _ in ds_eval.as_numpy_iterator():
        x_eval.append(x[0])
    x_eval = np.concatenate(x_eval, axis=0)
    for x, _ in ds_test.as_numpy_iterator():
        x_test.append(x[0])
    x_test = np.concatenate(x_test, axis=0)
    # Assert that duplicate of test data has the same shape:
    for x, _ in ds_test2:
        x_test2_shape += x[0].shape[0]
    assert x_test2_shape == x_test.shape[0]
    # Validate size of recovered numpy data sets:
    assert x_train.shape[0] + x_eval.shape[0] + x_test.shape[0] == test_estim.estimator.data.n_obs
    assert len(idx_train) + len(idx_eval) + len(idx_test) == test_estim.estimator.data.n_obs
    assert x_train.shape[0] == len(idx_train)
    assert x_eval.shape[0] == len(idx_eval)
    assert x_test.shape[0] == len(idx_test)
    # Assert that observations are unique within partition:
    assert np.all([
        np.sum(np.abs(x_train[[i], :] - x_train).sum(axis=1) == 0) == 1
        for i in range(x_train.shape[0])
    ])
    assert np.all([
        np.sum(np.abs(x_eval[[i], :] - x_eval).sum(axis=1) == 0) == 1
        for i in range(x_eval.shape[0])
    ])
    assert np.all([
        np.sum(np.abs(x_test[[i], :] - x_test).sum(axis=1) == 0) == 1
        for i in range(x_test.shape[0])
    ])
    # Assert that observations are not replicated across partitions:
    assert not np.any([
        np.any(np.abs(x_train[[i], :] - x_eval).sum(axis=1) == 0)
        for i in range(x_train.shape[0])
    ])
    assert not np.any([
        np.any(np.abs(x_train[[i], :] - x_test).sum(axis=1) == 0)
        for i in range(x_train.shape[0])
    ])
    assert not np.any([
        np.any(np.abs(x_eval[[i], :] - x_test).sum(axis=1) == 0)
        for i in range(x_eval.shape[0])
    ])



from typing import Union

from sfaira.data.dataloaders.base.dataset_group import DatasetGroup, DatasetSuperGroup
from sfaira.data.dataloaders.databases.cellxgene import DatasetSuperGroupCellxgene


class DatasetSuperGroupDatabases(DatasetSuperGroup):

    def __init__(
            self,
            data_path: Union[str, None] = None,
            meta_path: Union[str, None] = None,
            cache_path: Union[str, None] = None,
            cache_metadata: bool = False,
    ):
        dataset_super_groups = [
            DatasetSuperGroupCellxgene(
                data_path=data_path,
                meta_path=meta_path,
                cache_path=cache_path,
                cache_metadata=cache_metadata,
            ),
        ]
        super().__init__(dataset_groups=dataset_super_groups)



import anndata
import scipy.sparse
import numpy as np
import os
import pandas as pd
import pathlib

from sfaira.data.store.stores.multi import StoresAnndata
from sfaira.versions.genomes import GenomeContainer

from sfaira.unit_tests.directories import DIR_DATA_LOADERS_CACHE, DIR_DATA_LOADERS_STORE_DAO, \
    DIR_DATA_LOADERS_STORE_H5AD, save_delete
from .consts import RELEASE_HUMAN, RELEASE_MOUSE
from .loaders import DatasetSuperGroupMock

MATCH_TO_RELEASE = {"Homo sapiens": RELEASE_HUMAN,
                    "Mus musculus": RELEASE_MOUSE}


def _create_adata(celltypes, ncells, ngenes, assembly) -> anndata.AnnData:
    """
    Usesd by mock data loaders.
    """
    gc = GenomeContainer(organism=" ".join(assembly.split(".")[0].split("_")), release=assembly.split(".")[-1])
    gc.set(biotype="protein_coding")
    genes = gc.ensembl[:ngenes]
    x = scipy.sparse.csc_matrix(np.random.randint(low=0, high=100, size=(ncells, ngenes)))
    var = pd.DataFrame(index=genes)
    obs = pd.DataFrame({}, index=["cell_" + str(i) for i in range(ncells)])
    if len(celltypes) > 0:
        obs["free_annotation"] = [celltypes[i] for i in np.random.choice(len(celltypes), size=ncells, replace=True)]
    # Create random embedding
    obsm = {"X_umap": np.random.random(size=(ncells, 2))}
    adata = anndata.AnnData(X=x, obs=obs, obsm=obsm, var=var)
    return adata


def _load_script(dsg, rewrite: bool, match_to_release):
    dsg.load(allow_caching=True, load_raw=rewrite)
    dsg.streamline_features(remove_gene_version=True, match_to_release=match_to_release)
    dsg.streamline_metadata(schema="sfaira", clean_obs=True, clean_var=True, clean_uns=True, clean_obs_names=True)
    return dsg


class PrepareData:
    CLS_DSG = DatasetSuperGroupMock

    def prepare_dsg(self, rewrite: bool = False, load: bool = True, match_to_release=None):
        """
        Prepares data set super group of mock data and returns instance.

        Use this do testing involving a data set group.
        """
        # Make sure cache exists:
        if not os.path.exists(DIR_DATA_LOADERS_CACHE):
            pathlib.Path(DIR_DATA_LOADERS_CACHE).mkdir(parents=True, exist_ok=True)
        dsg = self.CLS_DSG()
        if match_to_release is None:
            match_to_release = MATCH_TO_RELEASE
        if load:
            dsg = _load_script(dsg=dsg, rewrite=rewrite, match_to_release=match_to_release)
        return dsg

    def prepare_store_anndata(self, match_to_reference=None) -> StoresAnndata:
        dsg = self.prepare_dsg(load=True, match_to_release=match_to_reference)
        store = StoresAnndata(adatas=dsg.adata_ls)
        return store

    def prepare_store(self, store_format: str, rewrite: bool = False, rewrite_store: bool = False,
                      match_to_reference=None) -> str:
        """
        Prepares mock data store and returns path to store.

        Use this do testing involving a data set store.
        """
        dir_store_formatted = {
            "dao": DIR_DATA_LOADERS_STORE_DAO,
            "h5ad": DIR_DATA_LOADERS_STORE_H5AD,
        }[store_format]
        if not os.path.exists(dir_store_formatted):
            pathlib.Path(dir_store_formatted).mkdir(parents=True, exist_ok=True)
        dsg = self.prepare_dsg(rewrite=rewrite, load=False, match_to_release=match_to_reference)
        for k, ds in dsg.datasets.items():
            print(k)
            if store_format == "dao":
                compression_kwargs = {"compressor": "default", "overwrite": True, "order": "C"}
            else:
                compression_kwargs = {}
            if store_format == "dao":
                anticipated_fn = os.path.join(dir_store_formatted, ds.doi_cleaned_id)
            elif store_format == "h5ad":
                anticipated_fn = os.path.join(dir_store_formatted, ds.doi_cleaned_id + ".h5ad")
            else:
                assert False
            if rewrite_store and os.path.exists(anticipated_fn):
                # Can't write if h5ad already exists.
                # Delete store to writing if forced.
                save_delete(anticipated_fn)
            # Only rewrite if necessary
            if rewrite_store or not os.path.exists(anticipated_fn):
                ds = _load_script(dsg=ds, rewrite=rewrite, match_to_release=MATCH_TO_RELEASE)
                ds.write_distributed_store(dir_cache=dir_store_formatted, store_format=store_format, dense=True,
                                           chunks=128, compression_kwargs=compression_kwargs)
        return dir_store_formatted



import abc
import networkx
import numpy as np
import obonet
import os
import owlready2
import pickle
import requests
from functools import lru_cache
from typing import Dict, List, Tuple, Union


from sfaira import settings

"""
Ontology managament classes.

We consider any structured collection of meta data identifiers an ontology and define classes to interact with such
data here.

- All classes inherit from Ontology()
- Onotlogies can be read as follows:
    - from string lists which are typically hardcoded in sfaira (OntologyList),
    - from .obo files which are emitted by obofoundry for example (OntologyObo)),
    - ToDo from .owl files which are emitted from EBI for example (OntologyOwl)),
    - from the EBI web API via direct queries (OntologyEbi)).

ToDo explain usage of ontology extension.
"""


def cached_load_file(url, ontology_cache_dir, ontology_cache_fn, recache: bool = False):
    if os.name == "nt":  # if running on windows, do not download obo file, but rather pass url directly to obonet
        # TODO add caching option.
        obofile = url
    else:
        ontology_cache_dir = os.path.join(settings.cachedir_ontologies, ontology_cache_dir)
        obofile = os.path.join(ontology_cache_dir, ontology_cache_fn)
        # Download if necessary:
        if not os.path.exists(obofile) or recache:
            os.makedirs(name=ontology_cache_dir, exist_ok=True)

            def download_file():
                print(f"Downloading: {ontology_cache_fn} to {ontology_cache_dir}")
                if not os.path.exists(ontology_cache_dir):
                    os.makedirs(ontology_cache_dir)
                r = requests.get(url, allow_redirects=True)
                # if url.startswith("https://raw.githubusercontent.com"):
                #     open(obofile, 'wb').write(r.text)
                # else:
                open(obofile, 'wb').write(r.content)

            download_file()
    return obofile


def cached_load_ebi(ontology_cache_dir, ontology_cache_fn, recache: bool = False) -> (networkx.MultiDiGraph, os.PathLike):
    """
    Load pickled graph object if available.

    :param ontology_cache_dir:
    :param ontology_cache_fn:
    :param recache:
    :return:
    """
    ontology_cache_dir = os.path.join(settings.cachedir_ontologies, ontology_cache_dir)
    picklefile = os.path.join(ontology_cache_dir, ontology_cache_fn)
    if os.path.isfile(picklefile) and not recache:
        with open(picklefile, 'rb') as f:
            graph = pickle.load(f)
    else:
        os.makedirs(name=ontology_cache_dir, exist_ok=True)
        graph = None
    return graph, picklefile


class Ontology:
    leaves: List[str]

    @abc.abstractmethod
    def node_names(self):
        pass

    @abc.abstractmethod
    def map_node_suggestion(self, x: str, include_synonyms: bool = True, n_suggest: int = 10):
        """
        Map free text node name to ontology node names via fuzzy string matching.

        :param x: Free text node label which is to be matched to ontology nodes.
        :param include_synonyms: Whether to search for meaches in synonyms field of node instances, too.
        :return List of proposed matches in ontology.
        """
        pass

    def is_node(self, x: str):
        return x in self.node_names

    def validate_node(self, x: str):
        if not self.is_node(x=x):
            suggestions = self.map_node_suggestion(x=x, include_synonyms=False)
            raise ValueError(f"Node label {x} not found. Did you mean any of {suggestions}?")


class OntologyList(Ontology):
    """
    Basic unordered ontology container.

    Node IDs and names are the same.
    """
    nodes: list

    def __init__(
            self,
            terms: Union[List[Union[str, bool, int]]],
            **kwargs
    ):
        self.nodes = terms

    @property
    def node_names(self) -> List[str]:
        return self.nodes

    @property
    def node_ids(self) -> List[str]:
        return self.nodes

    @property
    def leaves(self) -> List[str]:
        return self.nodes

    @property
    def n_leaves(self) -> int:
        return len(self.nodes)

    def prepare_maps_to_leaves(
            self,
            include_self: bool = True
    ) -> Dict[str, np.ndarray]:
        """
        Precomputes all maps of nodes to their leave nodes.

        Note that for a list ontology, this maps each node to itself.

        :param include_self: whether to include node itself
        :return: Dictionary of index vectors of leave node matches for each node (key).
        """
        if include_self:
            return dict([(x, np.array([self.leaves.index(x)])) for x in self.leaves])
        else:
            return dict([(x, np.array([])) for x in self.leaves])

    def is_a_node_id(self, x: str) -> bool:
        return x in self.node_names

    def is_a_node_name(self, x: str) -> bool:
        return x in self.node_names

    @staticmethod
    def convert_to_id(x: Union[str, List[str]]) -> Union[str, List[str]]:
        return x

    @staticmethod
    def convert_to_name(x: Union[str, List[str]]) -> Union[str, List[str]]:
        return x

    def map_node_suggestion(self, x: str, include_synonyms: bool = True, n_suggest: int = 10):
        """
        Map free text node name to ontology node names via fuzzy string matching.

        :param x: Free text node label which is to be matched to ontology nodes.
        :param include_synonyms: Whether to search for meaches in synonyms field of node instances, too.
        :param n_suggest: number of suggestions returned
        :return List of proposed matches in ontology.
        """
        from fuzzywuzzy import fuzz
        scores = np.array([
            np.max([
                fuzz.ratio(x.lower(), y.lower())
            ])
            for y in self.node_names
        ])
        # Suggest top n_suggest hits by string match:
        return [self.node_names[i] for i in np.argsort(scores)[-n_suggest:]][::-1]

    def synonym_node_properties(self) -> List[str]:
        return []

    def is_a(self, query: str, reference: str) -> bool:
        """
        Checks if query node is reference node.

        Note that there is no notion of ancestors for list ontologies.

        :param query: Query node name. Node ID or name.
        :param reference: Reference node name. Node ID or name.
        :return: If query node is reference node or an ancestor thereof.
        """
        return query == reference

    def get_ancestors(self, node: str) -> List[str]:
        return []


class OntologyHierarchical(Ontology, abc.ABC):
    """
    Basic ordered ontology container
    """
    _graph: networkx.MultiDiGraph

    @property
    def graph(self) -> networkx.MultiDiGraph:
        return self._graph

    @graph.setter
    def graph(self, graph: networkx.MultiDiGraph):
        self._graph = graph
        self.get_ancestors.cache_clear()
        self.get_descendants.cache_clear()

    def _check_graph(self):
        if not networkx.is_directed_acyclic_graph(self.graph):
            print(f"Ontology {type(self)} is not a DAG, treat child-parent reasoning with care.")

    def __validate_node_ids(self, x: Union[str, List[str]]):
        if isinstance(x, str):
            x = [x]
        node_ids = self.node_ids
        for y in x:
            if y not in node_ids:
                raise ValueError(f"queried node id {y} is not in graph")

    def __validate_node_names(self, x: Union[str, List[str]]):
        if isinstance(x, str):
            x = [x]
        node_names = self.node_names
        for y in x:
            if y not in node_names:
                raise ValueError(f"queried node name {y} is not in graph")

    @property
    def nodes(self) -> List[Tuple[str, dict]]:
        return list(self.graph.nodes.items())

    @property
    def nodes_dict(self) -> dict:
        return dict(list(self.graph.nodes.items()))

    @property
    def node_names(self) -> List[str]:
        try:
            return [x["name"] for x in self.graph.nodes.values()]
        except KeyError as e:
            raise KeyError(f"KeyError '{e}' in {type(self)}")

    @property
    def node_ids(self) -> List[str]:
        return list(self.graph.nodes())

    def is_a_node_id(self, x: str) -> bool:
        return x in self.node_ids

    def is_a_node_name(self, x: str) -> bool:
        return x in self.node_names

    def convert_to_name(self, x: Union[str, List[str]]) -> Union[str, List[str]]:
        was_str = isinstance(x, str)
        if was_str:
            x = [x]
        if self.is_a_node_id(x[0]):
            self.__validate_node_ids(x=x)
            x = [
                [v["name"] for k, v in self.graph.nodes.items() if k == z][0]
                for z in x
            ]
        elif self.is_a_node_name(x[0]):
            self.__validate_node_names(x=x)
        else:
            raise ValueError(f"node {x[0]} not recognized")
        self.__validate_node_names(x=x)
        if was_str:
            return x[0]
        else:
            return x

    def convert_to_id(self, x: Union[str, List[str]]) -> Union[str, List[str]]:
        was_str = isinstance(x, str)
        if was_str:
            x = [x]
        if self.is_a_node_id(x[0]):
            self.__validate_node_ids(x=x)
        elif self.is_a_node_name(x[0]):
            self.__validate_node_names(x=x)
            x = [
                [k for k, v in self.graph.nodes.items() if v["name"] == z][0]
                for z in x
            ]
        else:
            raise ValueError(f"node {x[0]} not recognized")
        self.__validate_node_ids(x=x)
        if was_str:
            return x[0]
        else:
            return x

    @lru_cache(maxsize=None)
    def __convert_to_id_cached(self, x: str) -> str:
        return self.convert_to_id(x)

    @property
    def leaves(self) -> List[str]:
        return [x for x in self.graph.nodes() if self.graph.in_degree(x) == 0]

    @leaves.setter
    def leaves(self, x: List[str]):
        """
        Sets new leaf-space for graph.

        This clips nodes that are not upstream of defined leaves.
        :param x: New set of leaves nodes, identified as IDs.
        """
        x = self.convert_to_id(x)
        nodes_to_remove = []
        for y in self.graph.nodes():
            if not np.any([self.is_a(query=z, reference=y, convert_to_id=False) for z in x]):
                nodes_to_remove.append(y)
        self.graph.remove_nodes_from(nodes_to_remove)

    @property
    def n_leaves(self) -> int:
        return len(self.leaves)

    def get_effective_leaves(self, x: List[str]) -> List[str]:
        """
        Get effective leaves in ontology given set of observed nodes.

        The effective leaves are the minimal set of nodes such that all nodes in x are ancestors of this set, ie the
        observed nodes which represent leaves of a sub-DAG of the ontology DAG, which captures all observed nodes.

        :param x: Observed node IDs.
        :return: Effective leaves.
        """
        if isinstance(x, str):
            x = [x]
        if isinstance(x, np.ndarray):
            x = x.tolist()
        assert isinstance(x, list), "supply either list or str to get_effective_leaves"
        if len(x) == 0:
            raise ValueError("x was empty list, get_effective_leaves cannot be called on empty list")
        x = list(np.unique(x))
        x = self.convert_to_id(x=x)
        leaves = []
        for y in x:
            if not np.any([self.is_a(query=z, reference=y, convert_to_id=False) for z in list(set(x) - {y})]):
                leaves.append(y)
        return leaves

    @lru_cache(maxsize=None)
    def get_ancestors(self, node: str) -> List[str]:
        node = self.__convert_to_id_cached(node)
        return list(networkx.ancestors(self.graph, node))

    @lru_cache(maxsize=None)
    def get_descendants(self, node: str) -> List[str]:
        node = self.__convert_to_id_cached(node)
        return list(networkx.descendants(self.graph, node))

    def is_a(self, query: str, reference: str, convert_to_id: bool = True) -> bool:
        """
        Checks if query node is reference node or an ancestor thereof.

        :param query: Query node name. Node ID or name.
        :param reference: Reference node name. Node ID or name.
        :param convert_to_id: Whether to call self.convert_to_id on `query` and `reference` input arguments
        :return: If query node is reference node or an ancestor thereof.
        """
        if convert_to_id:
            query = self.__convert_to_id_cached(query)
            reference = self.__convert_to_id_cached(reference)
        return query in self.get_ancestors(node=reference) or query == reference

    def map_to_leaves(
            self,
            node: str,
            return_type: str = "ids",
            include_self: bool = True
    ) -> Union[List[str], np.ndarray]:
        """
        Map a given node to leave nodes.

        :param node: Node(s) to map as symbol(s) or ID(s).
        :param return_type:

            "ids": IDs of mapped leave nodes
            "idx": indicies in leave note list of mapped leave nodes
        :param include_self: DEPRECEATED.
        :return:
        """
        node = self.__convert_to_id_cached(node)
        ancestors = self.get_ancestors(node)
        # Add node itself to list of ancestors.
        ancestors = ancestors + [node]
        leaves = self.leaves
        if return_type == "ids":
            return [x for x in leaves if x in ancestors]
        elif return_type == "idx":
            return np.sort([i for i, x in enumerate(leaves) if x in ancestors])
        else:
            raise ValueError(f"return_type {return_type} not recognized")

    def prepare_maps_to_leaves(
            self,
            include_self: bool = True
    ) -> Dict[str, np.ndarray]:
        """
        Precomputes all maps of nodes to their leave nodes.

        :param include_self: whether to include node itself
        :return: Dictionary of index vectors of leave node matches for each node (key).
        """
        nodes = self.node_ids
        maps = {}
        import time
        t0 = time.time()
        for x in nodes:
            maps[x] = self.map_to_leaves(node=x, return_type="idx", include_self=include_self)
        print(f"time for precomputing ancestors: {time.time()-t0}")
        return maps

    def reset_root(self, root: str):
        new_nodes = [self.convert_to_id(x=root)] + self.get_ancestors(node=root)
        self.graph = self.graph.subgraph(nodes=new_nodes)

    @abc.abstractmethod
    def synonym_node_properties(self) -> List[str]:
        pass


class OntologyEbi(OntologyHierarchical):
    """
    Recursively assembles ontology by querying EBI web interface.

    Not recommended for large ontologies because of the iterative query of the web API.
    """

    def __init__(
            self,
            ontology: str,
            root_term: str,
            additional_terms: dict,
            additional_edges: List[Tuple[str, str]],
            ontology_cache_fn: str,
            recache: bool,
            **kwargs
    ):
        # Note on base URL: EBI OLS points to different resources depending on the ontology used, this needs to be
        # accounted for here.
        if ontology == "hancestro":
            base_url = f"https://www.ebi.ac.uk/ols/api/ontologies/{ontology}/terms/" \
                       f"http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252F"
        elif ontology == "efo":
            base_url = f"https://www.ebi.ac.uk/ols/api/ontologies/{ontology}/terms/" \
                       f"http%253A%252F%252Fwww.ebi.ac.uk%252F{ontology}%252F"
        else:
            assert False

        def get_url_self(iri):
            return f"{base_url}{iri}"

        def get_url_children(iri):
            return f"{base_url}{iri}/children"

        def get_iri_from_node(x):
            return x["iri"].split("/")[-1]

        def get_id_from_iri(x):
            x = ":".join(x.split("_"))
            return x

        def get_id_from_node(x):
            x = get_iri_from_node(x)
            x = get_id_from_iri(x)
            return x

        def recursive_search(iri):
            """
            This function queries all nodes that are children of a given node at one time. This is faster than querying
            the characteristics of each node separately but leads to slightly awkward code, the root node has to be
            queried separately for example below.

            :param iri: Root node IRI.
            :return: Tuple of

                - nodes (dictionaries of node ID and node values) and
                - edges (node ID of parent and child).
            """
            terms_children = requests.get(get_url_children(iri=iri)).json()["_embedded"]["terms"]
            nodes_new = {}
            edges_new = []
            direct_children = []
            k_self = get_id_from_iri(iri)
            # Define root node if this is the first iteration, this node is otherwise not defined through values.
            if k_self == ":".join(root_term.split("_")):
                terms_self = requests.get(get_url_self(iri=iri)).json()
                nodes_new[k_self] = {
                    "name": terms_self["label"],
                    "description": terms_self["description"],
                    "synonyms": terms_self["synonyms"],
                    "has_children": terms_self["has_children"],
                }
            for c in terms_children:
                k_c = get_id_from_node(c)
                nodes_new[k_c] = {
                    "name": c["label"],
                    "description": c["description"],
                    "synonyms": c["synonyms"],
                    "has_children": c["has_children"],
                }
                direct_children.append(k_c)
                if c["has_children"]:
                    nodes_x, edges_x = recursive_search(iri=get_iri_from_node(c))
                    nodes_new.update(nodes_x)
                    # Update nested edges of between children:
                    edges_new.extend(edges_x)
            # Update edges to children:
            edges_new.extend([(k_self, k_c) for k_c in direct_children])
            return nodes_new, edges_new

        graph, picklefile = cached_load_ebi(ontology_cache_dir=ontology, ontology_cache_fn=ontology_cache_fn,
                                            recache=recache)
        if graph is None:
            self.graph = networkx.MultiDiGraph()
            nodes, edges = recursive_search(iri=root_term)
            nodes.update(additional_terms)
            edges.extend(additional_edges)
            for k, v in nodes.items():
                self.graph.add_node(node_for_adding=k, **v)
            for x in edges:
                parent, child = x
                self.graph.add_edge(child, parent)
            with open(picklefile, 'wb') as f:
                pickle.dump(obj=self.graph, file=f)
        else:
            self.graph = graph

    def map_node_suggestion(self, x: str, include_synonyms: bool = True, n_suggest: int = 10):
        """
        Map free text node name to ontology node names via fuzzy string matching.

        :param x: Free text node label which is to be matched to ontology nodes.
        :param include_synonyms: Whether to search for meaches in synonyms field of node instances, too.
        :return List of proposed matches in ontology.
        """
        from fuzzywuzzy import fuzz
        scores = np.array([
            np.max(
                [
                    fuzz.partial_ratio(x.lower(), v["name"].lower())
                ] + [
                    fuzz.partial_ratio(x.lower(), yyy.lower())
                    for yy in self.synonym_node_properties if yy in v.keys() for yyy in v[yy]
                ]
            ) if include_synonyms else
            np.max([
                fuzz.partial_ratio(x.lower(), v["name"].lower())
            ])
            for k, v in self.graph.nodes.items()
        ])
        # Suggest top n_suggest hits by string match:
        return [self.node_names[i] for i in np.argsort(scores)[-n_suggest:]][::-1]

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonyms"]


class OntologyOwl(OntologyHierarchical, abc.ABC):

    onto_owl = owlready2.Ontology

    def __init__(
            self,
            owl: str,
            **kwargs
    ):
        # self.onto_owl = owlready2.get_ontology(owl)
        # self.onto_owl.load()
        self.graph = None


class OntologyObo(OntologyHierarchical, abc.ABC):

    def __init__(
            self,
            obo: str,
            **kwargs
    ):
        self.graph = obonet.read_obo(obo, ignore_obsolete=True)

    def map_node_suggestion(self, x: str, include_synonyms: bool = True, n_suggest: int = 10):
        """
        Map free text node name to ontology node names via fuzzy string matching.

        :param x: Free text node label which is to be matched to ontology nodes.
        :param include_synonyms: Whether to search for meaches in synonyms field of node instances, too.
        :return List of proposed matches in ontology.
        """
        from fuzzywuzzy import fuzz
        scores = np.array([
            np.max(
                [
                    fuzz.ratio(x.lower().strip("'").strip("\""), y[1]["name"].lower())
                ] + [
                    fuzz.ratio(x.lower().strip("'").strip("\"").strip("]").strip("["), yyy.lower())
                    for yy in self.synonym_node_properties if yy in y[1].keys() for yyy in y[1][yy]
                ]
            ) if "synonym" in y[1].keys() and include_synonyms else
            np.max([
                fuzz.ratio(x.lower().strip("'").strip("\""), y[1]["name"].lower())
            ])
            for y in self.nodes
        ])
        # Suggest top n_suggest hits by string match:
        return [self.nodes[i][1]["name"] for i in np.argsort(scores)[-n_suggest:]][::-1]


class OntologyExtendedObo(OntologyObo):
    """
    Basic .obo ontology extended by additional nodes and edges without breaking DAG.
    """

    def __init__(self, obo, **kwargs):
        super().__init__(obo=obo, **kwargs)

    def add_extension(self, dict_ontology: Dict[str, List[Dict[str, dict]]]):
        """
        Extend ontology by additional edges and nodes defined in a dictionary.

        Checks that DAG is not broken after graph assembly.

        :param dict_ontology: Dictionary of nodes and edges to add to ontology. Parsing:

            - keys: parent nodes (which must be in ontology)
            - values: children nodes (which can be in ontology), must be given as a dictionary in which keys are
                ontology IDs and values are node values..
                If these are in the ontology, an edge is added, otherwise, an edge and the node are added.
        :return:
        """
        for k, v in dict_ontology.items():
            assert isinstance(v, dict), "dictionary values should be dictionaries"
            # Check that parent node is present:
            if k not in self.node_ids:
                raise ValueError(f"key {k} was not in reference ontology")
            # Check if edge is added only, or edge and node.
            for child_node_k, child_node_v in v.items():
                if child_node_k not in self.node_ids:  # Add node
                    self.graph.add_node(node_for_adding=child_node_k, **child_node_v)
                # Add edge.
                self.graph.add_edge(k, child_node_k)
        # Check that DAG was not broken:
        self._check_graph()

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym"]


class OntologyUberon(OntologyExtendedObo):

    def __init__(
            self,
            branch: str,
            recache: bool = False,
            **kwargs
    ):
        obofile = cached_load_file(
            url=f"https://raw.githubusercontent.com/obophenotype/uberon/{branch}/composite-vertebrate-basic.obo",
            ontology_cache_dir="uberon",
            ontology_cache_fn=f"uberon_{branch}.obo",
            recache=recache,
        )
        super().__init__(obo=obofile)

        # Clean up nodes:
        nodes_to_delete = []
        for k, v in self.graph.nodes.items():
            # Only retain nodes which are  "anatomical collection" 'UBERON:0034925':
            # ToDo this seems to narrow, need to check if we need to constrain the nodes we use.
            if "name" not in v.keys():
                nodes_to_delete.append(k)
        for k in nodes_to_delete:
            self.graph.remove_node(k)

        # Clean up edges:
        # The graph object can hold different types of edges,
        # and multiple types are loaded from the obo, not all of which are relevant for us:
        # All edge types (based on previous download, assert below that this is not extended):
        edge_types = [
            'aboral_to',
            'adjacent_to',
            'ambiguous_for_taxon',
            'anastomoses_with',
            'anterior_to',
            'anteriorly_connected_to',
            'attaches_to',
            'attaches_to_part_of',
            'bounding_layer_of',
            'branching_part_of',
            'capable_of',
            'capable_of_part_of',
            'channel_for',
            'channels_from',
            'channels_into',
            'composed_primarily_of',
            'conduit_for',
            'confers_advantage_in',
            'connected_to',
            'connects',
            'contains',
            'contains_process',
            'continuous_with',
            'contributes_to_morphology_of',
            'deep_to',
            'developmentally_induced_by',
            'developmentally_replaces',
            'develops_from',  # developmental DAG -> include because it reflects the developmental hierarchy
            'develops_from_part_of',  # developmental DAG -> include because it reflects the developmental hierarchy
            'develops_in',
            'directly_develops_from',  # developmental DAG -> include because it reflects the developmental hierarchy
            'distal_to',
            'distally_connected_to',
            'distalmost_part_of',
            'dorsal_to',
            'drains',
            'dubious_for_taxon',
            'ends',
            'ends_with',
            'existence_ends_during',
            'existence_ends_during_or_before',
            'existence_ends_with',
            'existence_starts_and_ends_during',
            'existence_starts_during',
            'existence_starts_during_or_after',
            'existence_starts_with',
            'extends_fibers_into',
            'filtered_through',
            'functionally_related_to',
            'has_boundary',
            'has_component',
            'has_developmental_contribution_from',
            'has_fused_element',
            'has_member',
            'has_muscle_antagonist',
            'has_muscle_insertion',
            'has_muscle_origin',
            'has_part',
            'has_potential_to_develop_into',
            'has_potential_to_developmentally_contribute_to',
            'has_quality',
            'has_skeleton',
            'immediate_transformation_of',
            'immediately_anterior_to',
            'immediately_deep_to',
            'immediately_posterior_to',
            'immediately_preceded_by',
            'immediately_superficial_to',
            'in_anterior_side_of',
            'in_central_side_of',
            'in_deep_part_of',
            'in_distal_side_of',
            'in_dorsal_side_of',
            'in_innermost_side_of',
            'in_lateral_side_of',
            'in_left_side_of',
            'in_outermost_side_of',
            'in_posterior_side_of',
            'in_proximal_side_of',
            'in_right_side_of',
            'in_superficial_part_of',
            'in_taxon',
            'in_ventral_side_of',
            'indirectly_supplies',
            'innervated_by',
            'innervates',
            'input_of',
            'intersects_midsagittal_plane_of',
            'is_a',  # term DAG -> include because it connect conceptual tissue groups
            'layer_part_of',
            'located_in',  # anatomic DAG -> include because it reflects the anatomic coarseness / hierarchy
            'location_of',
            'lumen_of',
            'luminal_space_of',
            'negatively_regulates',
            'never_in_taxon',
            'occurs_in',
            'only_in_taxon',
            'output_of',
            'overlaps',
            'part_of',  # anatomic DAG -> include because it reflects the anatomic coarseness / hierarchy
            'participates_in',
            'positively_regulates',
            'postaxialmost_part_of',
            'posterior_to',
            'posteriorly_connected_to',
            'preaxialmost_part_of',
            'preceded_by',
            'precedes',
            'present_in_taxon',
            'produced_by',
            'produces',
            'protects',
            'proximal_to',
            'proximally_connected_to',
            'proximalmost_part_of',
            'regulates',
            'seeAlso',
            'serially_homologous_to',
            'sexually_homologous_to',
            'simultaneous_with',
            'site_of',
            'skeleton_of',
            'starts',
            'starts_with',
            'subdivision_of',
            'superficial_to',
            'supplies',
            'surrounded_by',
            'surrounds',
            'synapsed_by',
            'transformation_of',
            'tributary_of',
            'trunk_part_of',
            'ventral_to'
        ]
        edges_to_delete = []
        for i, x in enumerate(self.graph.edges):
            if x[2] not in edge_types:
                print(f"NON-CRITICAL WARNING: uberon edge type {x[2]} not in reference list yet")
            if x[2] not in [
                "develops_from",
                'develops_from_part_of',
                'directly_develops_from',
                "is_a",
                "located_in",
                "part_of",
            ]:
                edges_to_delete.append((x[0], x[1]))
        for x in edges_to_delete:
            self.graph.remove_edge(u=x[0], v=x[1])
        self._check_graph()

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym", "latin term", "has relational adjective"]


class OntologyUberonLifecyclestage(OntologyUberon):

    """
    Subset of UBERON for generic life cycle stages that can be used for organism not covered by specific developmental
    ontologies.
    """

    def __init__(
            self,
            branch: str,
            recache: bool = False,
            **kwargs
    ):
        super().__init__(branch=branch, recache=recache, **kwargs)
        self.reset_root(root="UBERON:0000105")


class OntologyCl(OntologyExtendedObo):

    def __init__(
            self,
            branch: str,
            use_developmental_relationships: bool = False,
            recache: bool = False,
            **kwargs
    ):
        """

        Developmental edges are not desired in all interactions with this ontology, double-negative thymocytes are for
        example not an intuitive parent node for a fine grained T cell label in a non-thymic tissue.
        :param branch:
        :param use_developmental_relationships: Whether to keep developmental relationships.
        :param kwargs:
        """
        obofile = cached_load_file(
            url=f"https://raw.github.com/obophenotype/cell-ontology/{branch}/cl.obo",
            ontology_cache_dir="cl",
            ontology_cache_fn=f"cl_{branch}.obo",
            recache=recache,
        )
        super().__init__(obo=obofile)

        # Clean up nodes:
        nodes_to_delete = []
        for k, v in self.graph.nodes.items():
            # Some terms are not associated with the namespace cell but are cell types,
            # we identify these based on their ID nomenclature here.
            if ("namespace" in v.keys() and v["namespace"] not in ["cell", "cl"]) or \
                    ("namespace" not in v.keys() and str(k)[:2] != "CL"):
                nodes_to_delete.append(k)
            elif "name" not in v.keys():
                nodes_to_delete.append(k)
        for k in nodes_to_delete:
            self.graph.remove_node(k)

        # Clean up edges:
        # The graph object can hold different types of edges,
        # and multiple types are loaded from the obo, not all of which are relevant for us:
        # All edge types (based on previous download, assert below that this is not extended):
        edge_types = [
            'is_a',  # nomenclature DAG -> include because of annotation coarseness differences
            'derives_from',
            'develops_from',  # developmental DAG -> include because of developmental differences
            'has_part',  # ?
            'develops_into',  # inverse developmental DAG -> do not include
            'part_of',
            'RO:0002120',  # ?
            'RO:0002103',  # ?
            'lacks_plasma_membrane_part',  # ?
        ]
        edges_to_delete = []
        if use_developmental_relationships:
            edges_allowed = ["is_a", "develops_from"]
        else:
            edges_allowed = ["is_a"]
        for i, x in enumerate(self.graph.edges):
            if x[2] not in edge_types:
                print(f"NON-CRITICAL WARNING: cl edge type {x[2]} not in reference list yet")
            if x[2] not in edges_allowed:
                edges_to_delete.append((x[0], x[1]))
        for x in edges_to_delete:
            self.graph.remove_edge(u=x[0], v=x[1])
        self._check_graph()

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym"]


class OntologyOboCustom(OntologyExtendedObo):

    def __init__(
            self,
            obo: str,
            **kwargs
    ):
        super().__init__(obo=obo, **kwargs)


# use OWL for OntologyHancestro


class OntologyHsapdv(OntologyExtendedObo):

    def __init__(
            self,
            branch: str,
            recache: bool = False,
            **kwargs
    ):
        # Note on URL: berkeleybop is the ontology supported by cellxgene, GitHub seems to be older but versioned.
        obofile = cached_load_file(
            url="http://ontologies.berkeleybop.org/hsapdv.obo",
            # url=f"https://raw.githubusercontent.com/obophenotype/developmental-stage-ontologies/{branch}/src/hsapdv/hsapdv.obo",
            ontology_cache_dir="hsapdv",
            ontology_cache_fn=f"hsapdv_{branch}.obo",
            recache=recache,
        )
        super().__init__(obo=obofile)

        # Clean up nodes:
        nodes_to_delete = []
        for k, v in self.graph.nodes.items():
            if "name" not in v.keys():
                nodes_to_delete.append(k)
        for k in nodes_to_delete:
            self.graph.remove_node(k)

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym"]


class OntologyMmusdv(OntologyExtendedObo):

    def __init__(
            self,
            branch: str,
            recache: bool = False,
            **kwargs
    ):
        # URL for releases, not used here yet because versioning with respect to releases below is not consistent yet.
        # url=f"https://raw.githubusercontent.com/obophenotype/developmental-stage-ontologies/{branch}/src/mmusdv/mmusdv.obo"
        obofile = cached_load_file(
            url="http://ontologies.berkeleybop.org/mmusdv.obo",
            ontology_cache_dir="mmusdv",
            ontology_cache_fn="mmusdv.obo",
            recache=recache,
        )
        super().__init__(obo=obofile)

        # Clean up nodes:
        nodes_to_delete = []
        for k, v in self.graph.nodes.items():
            if "name" not in v.keys():
                nodes_to_delete.append(k)
        for k in nodes_to_delete:
            self.graph.remove_node(k)

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym"]


class OntologyMondo(OntologyExtendedObo):

    def __init__(
            self,
            branch: str,
            recache: bool = False,
            **kwargs
    ):
        # Latest release also available from url="http://purl.obolibrary.org/obo/mondo.obo".
        obofile = cached_load_file(
            url=f"https://raw.githubusercontent.com/monarch-initiative/mondo/{branch}/mondo-lastbuild.obo",
            ontology_cache_dir="mondo",
            ontology_cache_fn=f"mondo_{branch}.obo",
            recache=recache,
        )
        super().__init__(obo=obofile)

        # Clean up nodes:
        nodes_to_delete = []
        for k, v in self.graph.nodes.items():
            if "name" not in v.keys():
                nodes_to_delete.append(k)
        for k in nodes_to_delete:
            self.graph.remove_node(k)

        # add healthy property
        # Add node "healthy" under root node "MONDO:0000001": "quality".
        # We use a PATO node for this label: PATO:0000461.
        self.add_extension(dict_ontology={
            "MONDO:0000001": {
                "PATO:0000461": {"name": "healthy"}
            },
        })

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym"]


class OntologyCellosaurus(OntologyExtendedObo):

    def __init__(
            self,
            recache: bool = False,
            **kwargs
    ):
        obofile = cached_load_file(
            url="https://ftp.expasy.org/databases/cellosaurus/cellosaurus.obo",
            ontology_cache_dir="cellosaurus",
            ontology_cache_fn="cellosaurus.obo",
            recache=recache,
        )
        super().__init__(obo=obofile)

        # Clean up nodes:
        # edge_types = ["derived_from", "originate_from_same_individual_as"]
        nodes_to_delete = []
        for k, v in self.graph.nodes.items():
            if "name" not in v.keys():
                nodes_to_delete.append(k)
        for k in nodes_to_delete:
            self.graph.remove_node(k)

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym"]


class OntologyHancestro(OntologyEbi):

    """
    TODO move this to .owl backend once available.
    TODO root term: No term HANCESTRO_0001 ("Thing"?) accessible through EBI interface, because of that country-related
        higher order terms are not available as they are parallel to HANCESTRO_0004. Maybe fix with .owl backend?
    """

    def __init__(self, recache: bool = False):
        super().__init__(
            ontology="hancestro",
            root_term="HANCESTRO_0004",
            additional_terms={},
            additional_edges=[],
            ontology_cache_fn="hancestro.pickle",
            recache=recache,
        )


class OntologyTaxon(OntologyExtendedObo):

    """
    Note on ontology: The same repo also contains ncbitaxon.obs, the full ontology which is ~500MB large and
    takes multiple minutes to load. We are using a reduced version, taxslim, here.

    See also https://github.com/obophenotype/ncbitaxon/releases/download/{branch}/ncbitaxon.obo.
    """

    def __init__(
            self,
            branch: str,
            recache: bool = False,
            **kwargs
    ):
        obofile = cached_load_file(
            url=f"https://github.com/obophenotype/ncbitaxon/releases/download/{branch}/taxslim.obo",
            ontology_cache_dir="ncbitaxon",
            ontology_cache_fn=f"ncbitaxon_{branch}.obo",
            recache=recache,
        )
        super().__init__(obo=obofile)

        # Clean up nodes:
        nodes_to_delete = []
        for k, v in self.graph.nodes.items():
            if "name" not in v.keys():
                nodes_to_delete.append(k)
        for k in nodes_to_delete:
            self.graph.remove_node(k)

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym"]


class OntologyEfo(OntologyExtendedObo):

    def __init__(
            self,
            recache: bool = False,
            **kwargs
    ):
        obofile = cached_load_file(
            url="https://www.ebi.ac.uk/efo/efo.obo",
            ontology_cache_dir="efo",
            ontology_cache_fn="efo.obo",
            recache=recache,
        )
        super().__init__(obo=obofile)

        # Clean up nodes:
        nodes_to_delete = []
        for k, v in self.graph.nodes.items():
            if "name" not in v.keys():
                nodes_to_delete.append(k)
        for k in nodes_to_delete:
            self.graph.remove_node(k)

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym"]


class OntologySex(OntologyExtendedObo):

    """
    Sex is defined based on a subset of the PATO ontology.
    """

    def __init__(
            self,
            branch: str,
            recache: bool = False,
            **kwargs
    ):
        obofile = cached_load_file(
            url=f"https://raw.githubusercontent.com/pato-ontology/pato/{branch}/pato-base.obo",
            ontology_cache_dir="pato",
            ontology_cache_fn=f"pato_{branch}.obo",
            recache=recache,
        )
        super().__init__(obo=obofile)
        nodes_to_delete = []
        # Subset ontology: see also:
        # https://github.com/chanzuckerberg/single-cell-curation/blob/main/schema/2.0.0/schema.md#sex_ontology_term_id
        targets = self.get_ancestors("PATO:0001894")
        for k, v in self.graph.nodes.items():
            if k not in targets:
                nodes_to_delete.append(k)
        # Clean up nodes:
        for k, v in self.graph.nodes.items():
            if "name" not in v.keys():
                nodes_to_delete.append(k)
        nodes_to_delete = np.unique(nodes_to_delete)
        for k in nodes_to_delete:
            self.graph.remove_node(k)

    @property
    def synonym_node_properties(self) -> List[str]:
        return ["synonym"]



import numpy as np
from sfaira.consts.ontologies import DEFAULT_CL, DEFAULT_HSAPDV, DEFAULT_MONDO, DEFAULT_MMUSDV, DEFAULT_PATO, \
    DEFAULT_NCBITAXON, DEFAULT_UBERON
from sfaira.versions.metadata import OntologyUberon, OntologyCl, OntologyHancestro, OntologyHsapdv, OntologyMondo, \
    OntologyMmusdv, OntologyEfo, OntologyTaxon, OntologySex, OntologyUberonLifecyclestage

"""
OntologyCelltypes
"""


def test_cl_loading():
    """
    Tests if ontology can be initialised.
    """
    _ = OntologyCl(branch=DEFAULT_CL, recache=True)
    _ = OntologyCl(branch=DEFAULT_CL, recache=False)


def test_cl_is_a():
    """
    Tests if is-a relationships work correctly.
    """
    oc = OntologyCl(branch=DEFAULT_CL)
    assert oc.is_a(query="T cell", reference="lymphocyte")
    assert oc.is_a(query="lymphocyte", reference="lymphocyte")
    assert not oc.is_a(query="lymphocyte", reference="T cell")


def test_effective_leaves():
    """
    Tests if node sets can be mapped to effective leaf sets via `OntologyCelltypes.get_effective_leaves()`
    """
    oc = OntologyCl(branch=DEFAULT_CL)
    x = oc.get_effective_leaves(x=[
        "CD4-positive helper T cell", "lymphocyte", "stromal cell", "T cell", "T-helper 1 cell",
        "T-helper 17 cell"
    ])
    x = oc.convert_to_name(x)
    assert set(x) == {"stromal cell", "T-helper 1 cell", "T-helper 17 cell"}, x


def test_map_leaves():
    """
    Tests if nodes can be mapped to leave nodes in ontology.
    """
    oc = OntologyCl(branch=DEFAULT_CL)
    leaf_map_1 = oc.convert_to_name(oc.map_to_leaves(node="CD4-positive helper T cell", include_self=True))
    leaf_map_2 = oc.map_to_leaves(node="CD4-positive helper T cell", include_self=True, return_type="idx")
    assert len(leaf_map_1) == 7
    assert np.all(leaf_map_2 == np.sort([oc.convert_to_name(oc.leaves).index(x) for x in list(leaf_map_1)]))


def test_set_leaves():
    """
    Tests if ontology behaves correctly if leaf nodes were reset.
    """
    oc = OntologyCl(branch=DEFAULT_CL, use_developmental_relationships=False)
    targets = ["stromal cell", "T-helper 1 cell", "T-helper 17 cell"]
    oc.leaves = targets
    leaves = oc.convert_to_name(oc.leaves)
    assert set(leaves) == set(targets), leaves
    assert len(oc.node_ids) == 22
    assert np.all([x in oc.convert_to_name(oc.node_ids) for x in targets]), oc.convert_to_name(oc.node_ids)
    leaf_map_1 = oc.convert_to_name(oc.map_to_leaves(node="lymphocyte"))
    leaf_map_2 = oc.map_to_leaves(node="lymphocyte", return_type="idx")
    leaf_map_3 = oc.convert_to_name(oc.map_to_leaves(node="T-helper 1 cell"))
    leaf_map_4 = oc.map_to_leaves(node="T-helper 1 cell", return_type="idx")
    assert set(leaf_map_1) == {"T-helper 1 cell", "T-helper 17 cell"}
    assert np.all(leaf_map_2 == np.sort([oc.convert_to_name(oc.leaves).index(x) for x in list(leaf_map_1)]))
    assert set(leaf_map_3) == {"T-helper 1 cell"}
    assert np.all(leaf_map_4 == np.sort([oc.convert_to_name(oc.leaves).index(x) for x in list(leaf_map_3)]))


def test_reset_root():
    """
    Tests if root can be reset correctly.
    """
    oc = OntologyCl(branch=DEFAULT_CL, use_developmental_relationships=False)
    oc.reset_root(root="T cell")
    assert "T-helper 1 cell" in oc.node_names
    assert "T cell" in oc.node_names
    assert "lymphocyte" not in oc.node_names


"""
OntologyEfo
"""


def test_sclc_loading():
    """
    Tests if ontology can be initialised.
    """
    _ = OntologyEfo(recache=True)
    _ = OntologyEfo(recache=False)


def test_sclc_nodes():
    """
    Tests for presence and absence of a few commonly mistaken nodes.
    """
    sclc = OntologyEfo()
    assert "10x technology" in sclc.node_names
    assert "10x 3' v3" in sclc.node_names
    assert "Smart-like" in sclc.node_names
    assert "Smart-seq2" in sclc.node_names
    assert "sci-Plex" in sclc.node_names
    assert "single cell library construction" in sclc.node_names


def test_sclc_is_a():
    """
    Tests if is-a relationships work correctly.
    """
    sclc = OntologyEfo()
    assert sclc.is_a(query="10x 3' v3", reference="10x technology")
    assert sclc.is_a(query="10x 3' v3", reference="10x 3' transcription profiling")
    assert not sclc.is_a(query="10x technology", reference="10x 3' transcription profiling")
    assert sclc.is_a(query="10x 3' v3", reference="single cell library construction")
    assert sclc.is_a(query="sci-Plex", reference="single cell library construction")
    assert not sclc.is_a(query="sci-Plex", reference="10x technology")


"""
Hancestro
"""


def test_hancestro_loading():
    _ = OntologyHancestro(recache=True)
    _ = OntologyHancestro(recache=False)


"""
Hsapdv
"""


def test_hsapdv_loading():
    _ = OntologyHsapdv(branch=DEFAULT_HSAPDV, recache=True)
    _ = OntologyHsapdv(branch=DEFAULT_HSAPDV, recache=False)


"""
MONDO
"""


def test_mondo_loading():
    _ = OntologyMondo(branch=DEFAULT_MONDO, recache=True)
    _ = OntologyMondo(branch=DEFAULT_MONDO, recache=False)


"""
Mmusdv
"""


def test_mmusdv_loading():
    _ = OntologyMmusdv(branch=DEFAULT_MMUSDV, recache=True)
    _ = OntologyMmusdv(branch=DEFAULT_MMUSDV, recache=False)


"""
NCBI Taxon
"""


def test_taxon_loading():
    _ = OntologyTaxon(branch=DEFAULT_NCBITAXON, recache=True)
    _ = OntologyTaxon(branch=DEFAULT_NCBITAXON, recache=False)


"""
Sex
"""


def test_sex_loading():
    _ = OntologySex(branch=DEFAULT_PATO, recache=True)
    _ = OntologySex(branch=DEFAULT_PATO, recache=False)


"""
UBERON
"""


def test_uberon_loading():
    _ = OntologyUberon(branch=DEFAULT_UBERON, recache=True)
    _ = OntologyUberon(branch=DEFAULT_UBERON, recache=False)


def test_uberon_lcs_loading():
    _ = OntologyUberonLifecyclestage(branch=DEFAULT_UBERON, recache=False)


def test_uberon_subsetting():
    ou = OntologyUberon(branch=DEFAULT_UBERON)
    assert ou.is_a(query="lobe of lung", reference="lung")
    assert ou.is_a(query="lobe of lung", reference="lobe of lung")
    assert not ou.is_a(query="lung", reference="lobe of lung")

    assert ou.is_a(query="lobar bronchus", reference="lung")
    assert ou.is_a(query="lobar bronchus", reference="lobar bronchus")
    assert not ou.is_a(query="lung", reference="lobar bronchus")

    assert ou.is_a(query="adipose tissue of abdominal region", reference="adipose tissue")
    assert ou.is_a(query="adipose tissue", reference="adipose tissue")
    assert not ou.is_a(query="adipose tissue", reference="adipose tissue of abdominal region")



from sfaira.data.dataloaders.base.dataset import DatasetBase
from sfaira.data.dataloaders.base.dataset_group import DatasetGroup, DatasetGroupDirectoryOriented, DatasetSuperGroup
from sfaira.data.dataloaders.base.utils import clean_string



import anndata
import os
import numpy as np
import scipy.sparse

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://covid19.cog.sanger.ac.uk/james20.processed.h5ad"
        self.download_url_meta = None

        self.assay_sc_obs_key = "assay"
        self.author = "James"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41590-020-0602-z"
        self.individual_obs_key = "donor"
        self.layer_counts = "X"
        self.organ = "colon"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.state_exact = "healthy"
        self.year = 2020

        self.feature_id_var_key = "gene_ids"
        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "cell_type"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = os.path.join(data_dir, "james20.processed.h5ad")
    adata = anndata.read(fn)
    adata.X = np.expm1(adata.X)
    adata.X = adata.X.multiply(scipy.sparse.csc_matrix(adata.obs["n_counts"].values[:, None])).multiply(1 / 10000)
    # Assay maps are described here: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7212050/
    adata.obs["assay"] = [{
        '290b': "10x 3' transcription profiling",
        '298c': "10x 3' transcription profiling",
        '302c': "10x 3' transcription profiling",
        '390c': "10x 5' transcription profiling",
        '417c': "10x 5' transcription profiling",
    }[x] for x in adata.obs["donor"].values]
    return adata



import pydoc
import os
from typing import List, Union
from warnings import warn
from sfaira.data.dataloaders.base.dataset_group import DatasetSuperGroup, DatasetGroupDirectoryOriented


class DatasetSuperGroupLoaders(DatasetSuperGroup):

    dataset_groups: List[DatasetGroupDirectoryOriented]

    def __init__(
            self,
            data_path: Union[str, None] = None,
            meta_path: Union[str, None] = None,
            cache_path: Union[str, None] = None,
    ):
        """
        Class that sits on top of a directory of data set directories that each contain a data set group.

        :param file_base:
        :param dir_prefix: Prefix to sub-select directories by. Set to "" for no constraints.
        :param data_path:
        :param meta_path:
        :param cache_path:
        """
        # Directory choice hyperparamters:
        dir_prefix = "d"
        dir_exclude = []
        # Collect all data loaders from files in directory:
        dataset_groups = []
        cwd = os.path.dirname(__file__)
        for f in os.listdir(cwd):
            if os.path.isdir(os.path.join(cwd, f)):  # only directories
                if f[:len(dir_prefix)] == dir_prefix and f not in dir_exclude:  # Narrow down to data set directories
                    path_dsg = str(pydoc.locate(f"sfaira.data.dataloaders.loaders.{f}.FILE_PATH"))
                    if path_dsg is not None:
                        try:
                            dsg = DatasetGroupDirectoryOriented(
                                file_base=path_dsg,
                                data_path=data_path,
                                meta_path=meta_path,
                                cache_path=cache_path
                            )
                            dsg.collection_id = f
                            dataset_groups.append(dsg)
                        except IndexError as e:
                            raise IndexError(f"{e} for '{cwd}', '{f}', '{path_dsg}'")
                    else:
                        warn(f"DatasetGroupDirectoryOriented was None for {f}")
        super().__init__(dataset_groups=dataset_groups)



import os
import pandas as pd
import anndata

from sfaira.data import DatasetBase

SAMPLE_FNS = [
    "E-MTAB-6678.processed",
    "E-MTAB-6701.processed",
]


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = f"https://www.ebi.ac.uk/arrayexpress/files/{self.sample_fn.split('.')[0]}/" \
                                 f"{self.sample_fn}.1.zip"
        self.download_url_meta = f"https://www.ebi.ac.uk/arrayexpress/files/{self.sample_fn.split('.')[0]}/" \
                                 f"{self.sample_fn}.2.zip"

        self.assay_sc = "10x 3' v2" if self.sample_fn == "E-MTAB-6678.processed" else "Smart-seq2"
        self.author = "Ventotormo"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41586-018-0698-6"
        self.layer_counts = "X"
        self.organ = "placenta"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2018

        self.feature_id_var_key = "ensembl"
        self.feature_symbol_var_key = "names"
        self.feature_type = "rna"
        self.cell_type_obs_key = "annotation"

        self.set_dataset_id(idx=1)


def load(data_dir, sample_fn, **kwargs):
    fn = [
        os.path.join(data_dir, f"{sample_fn}.1.zip"),
        os.path.join(data_dir, f"{sample_fn}.2.zip"),
    ]
    adata = anndata.AnnData(pd.read_csv(fn[0], sep="\t", index_col="Gene").T)
    df = pd.read_csv(fn[1], sep="\t")
    for i in df.columns:
        adata.obs[i] = [df.loc[j][i] for j in adata.obs.index]

    adata.var["ensembl"] = [i.split("_")[1] for i in adata.var.index]
    adata.var["names"] = [i.split("_")[0] for i in adata.var.index]
    adata.var = adata.var.reset_index().reset_index().drop("index", axis=1)
    adata = adata[:, ~adata.var.index.isin(
        ["", "-1", "-10", "-11", "-2", "-3", "-4", "-5", "-6", "-7", "-8", "-9", "A.2", "A.3"])].copy()

    return adata




import anndata
import os
import pandas as pd
import scipy.io
import tarfile


def load(data_dir, sample_fn, **kwargs):
    fn = [os.path.join(data_dir, "GSE151511_RAW.tar"),
          os.path.join(data_dir, "41591_2020_1061_MOESM3_ESM.xlsx")]

    with tarfile.open(fn[0]) as tar:
        with tarfile.open(fileobj=tar.extractfile(sample_fn)) as sample:
            name = sample.getnames()[0]

            x = scipy.io.mmread(sample.extractfile(f'{name}/{name}_matrix.mtx')).T.tocsr()
            obs = pd.read_csv(sample.extractfile(f'{name}/{name}_barcodes.tsv'), header=None, sep="\t", index_col=0)
            obs.index.name = None
            var = pd.read_csv(sample.extractfile(f'{name}/{name}_genes.tsv'), header=None, sep="\t")
            var.columns = ["ensembl", "names"]
            var.index = var["ensembl"].values
            adata = anndata.AnnData(X=x, obs=obs, var=var)
    metadata = pd.read_excel(fn[1], skiprows=2, skipfooter=3)

    adata.obs['Sample ID'] = name
    adata.obs = adata.obs.reset_index().merge(metadata, on='Sample ID').set_index("index")
    adata.obs["Age"] = [str(x) for x in adata.obs["Age"].values]

    adata.obs = adata.obs.astype(str)
    return adata



from sfaira.data.dataloaders.databases.cellxgene.cellxgene_group import DatasetSuperGroupCellxgene, DatasetGroupCellxgene
from sfaira.data.dataloaders.databases.cellxgene.cellxgene_loader import Dataset



from sfaira.data.store.stores.base import StoreBase
from sfaira.data.store.stores.multi import StoreMultipleFeatureSpaceBase, StoresAnndata, \
    StoresDao, StoresH5ad
from sfaira.data.store.stores.single import StoreSingleFeatureSpace, StoreAnndata, \
    StoreDao



import anndata
import os
import numpy as np

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://covid19.cog.sanger.ac.uk/voigt19.processed.h5ad"
        self.download_url_meta = None

        self.assay_sc = "10x 3' v3"
        self.author = "Voigt"
        self.disease = "healthy"
        self.doi_journal = "10.1073/pnas.1914143116"
        self.layer_processed = "X"
        self.organ = "retina"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2019

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "CellType"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = os.path.join(data_dir, "voigt19.processed.h5ad")
    adata = anndata.read(fn)
    adata.X = np.expm1(adata.X)

    return adata



import anndata

from sfaira.unit_tests.data_for_tests.loaders.consts import ASSEMBLY_MOUSE, CELLTYPES
from sfaira.unit_tests.data_for_tests.loaders.utils import _create_adata


def load(data_dir, sample_fn, **kwargs) -> anndata.AnnData:
    ncells = 100
    ngenes = 70
    adata = _create_adata(celltypes=CELLTYPES[3:6], ncells=ncells, ngenes=ngenes,
                          assembly=ASSEMBLY_MOUSE)
    return adata
